import os
from functools import partial
from glob import glob
from pathlib import Path
from typing import List

from qtpy.QtCore import QPoint, Qt, Signal
from qtpy.QtGui import QDragEnterEvent, QDropEvent
from qtpy.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QFileDialog,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from PartSeg.common_backend.base_settings import BaseSettings
from PartSeg.common_gui.custom_load_dialog import SelectDirectoryDialog
from PartSeg.common_gui.main_window import OPEN_DIRECTORY
from PartSegCore.analysis.calculation_plan import MaskMapper

IO_BATCH_DIRECTORY = "io.batch_directory"


class AcceptFiles(QDialog):
    def __init__(self, files):
        super().__init__()
        self.ok = QPushButton("Add", self)
        self.ok.clicked.connect(self.accept)
        discard = QPushButton("Discard", self)
        discard.clicked.connect(self.close)
        self.files = QListWidget(self)
        self.files.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        for file_name in files:
            self.files.addItem(file_name)
        for i in range(self.files.count()):
            self.files.item(i).setSelected(True)
        self.ok.setDefault(True)
        self.ok.setAutoDefault(True)

        layout = QVBoxLayout()
        layout.addWidget(QLabel(f"Found {len(files)} files"))
        layout.addWidget(self.files)
        butt_layout = QHBoxLayout()
        butt_layout.addWidget(discard)
        butt_layout.addStretch()
        butt_layout.addWidget(self.ok)
        layout.addLayout(butt_layout)
        self.setLayout(layout)

    def selection_changed(self):
        if self.files.selectedItems().count() == 0:
            self.ok.setDisabled(True)
        else:
            self.ok.setEnabled(True)

    def get_files(self):
        return [str(item.text()) for item in self.files.selectedItems()]


class FileListItem(QListWidgetItem):
    def __init__(self, file_path):
        size = os.stat(file_path).st_size
        size = float(size) / (1024**2)
        super().__init__(f"{file_path:s} ({size:.2f} MB)")
        self.setTextAlignment(Qt.AlignmentFlag.AlignRight)
        self.file_path = file_path


class FileListWidget(QListWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)


class AddFiles(QWidget):
    """Docstring for AddFiles."""

    file_list_changed = Signal(set)

    def __init__(self, settings: BaseSettings, parent=None, btn_layout=QHBoxLayout):
        """TODO: to be defined1."""
        super().__init__(parent)
        self.mask_list: List[MaskMapper] = []
        self.settings = settings
        self.files_to_proceed = set()
        self.paths_input = QLineEdit(self)
        self.add_suffix_btn = QPushButton("*[0-9].tif", self)
        self.add_suffix_btn.clicked.connect(self.add_suffix_action)
        self.selected_files = FileListWidget(self)
        self.selected_files.itemSelectionChanged.connect(self.file_chosen)
        self.found_button = QPushButton("Find all", self)
        self.found_button.clicked.connect(self.find_all)
        self.select_files_button = QPushButton("Select files")
        self.select_dir_button = QPushButton("Select directory")
        self.select_files_button.clicked.connect(self.select_files)
        self.select_dir_button.clicked.connect(self.select_directory)
        self.delete_button = QPushButton("Remove file", self)
        self.delete_button.setDisabled(True)
        self.delete_button.clicked.connect(self.delete_element)
        self.clean_button = QPushButton("Remove all", self)
        self.clean_button.clicked.connect(self.clean)
        layout = QVBoxLayout()
        input_layout = QHBoxLayout()
        input_layout.addWidget(self.paths_input)
        input_layout.addWidget(self.add_suffix_btn)
        layout.addLayout(input_layout)
        select_layout = btn_layout()
        select_layout.addWidget(self.select_files_button)
        select_layout.addWidget(self.select_dir_button)
        select_layout.addWidget(self.found_button)
        select_layout.addStretch()
        select_layout.addWidget(self.clean_button)
        select_layout.addWidget(self.delete_button)
        layout.addLayout(select_layout)
        layout.addWidget(self.selected_files)
        self.setLayout(layout)
        self.setAcceptDrops(True)
        self.selected_files.customContextMenuRequested.connect(self.files_context_menu)

    def add_suffix_action(self):
        if self.paths_input.text() and not self.paths_input.text().endswith(self.add_suffix_btn.text()):
            self.paths_input.setText(os.path.join(self.paths_input.text(), self.add_suffix_btn.text()))

    def files_context_menu(self, point: QPoint):
        element = self.selected_files.itemAt(point)
        if element is None:
            return
        menu = QMenu()
        menu.addAction("Load image").triggered.connect(self._load_file)
        for mask_def in self.mask_list:
            menu.addAction(f"Load with mask '{mask_def.name}'").triggered.connect(
                partial(self._load_file_with_mask, mask_def)
            )
        menu.addAction("Delete").triggered.connect(self.delete_element)
        menu.exec_(self.selected_files.mapToGlobal(point))

    def _load_file(self):
        file_path = self.selected_files.item(self.selected_files.currentRow()).file_path
        self.settings._load_files_call([file_path])  # pylint: disable=protected-access

    def _load_file_with_mask(self, mask_mapper: MaskMapper):
        file_path = self.selected_files.item(self.selected_files.currentRow()).file_path
        mask_path = mask_mapper.get_mask_path(file_path)
        self.settings._load_files_call([file_path, mask_path])  # pylint: disable=protected-access

    def dragEnterEvent(self, event: QDragEnterEvent):  # pylint: disable=no-self-use
        if event.mimeData().hasFormat("text/plain"):
            event.acceptProposedAction()

    def dropEvent(self, event: QDropEvent):
        files_list = event.mimeData().text().split("\n")
        self.parse_drop_file_list(files_list)

    def parse_drop_file_list(self, files_list):
        res_list = []
        base_path = self.paths_input.text()
        is_broken_path = "*" in base_path or not os.path.isdir(base_path)

        for file_path in files_list:
            if os.path.isabs(file_path):
                res_list.append(file_path)
            elif is_broken_path:
                QMessageBox.warning(self, "Not proper path", f"Current path {base_path} is not a proper directory.")
                return
            else:
                res_list.append(os.path.join(base_path, file_path.strip()))
        if missed_files := [x for x in res_list if not os.path.exists(x)]:
            if len(missed_files) > 6:
                missed_files = missed_files[:6] + ["..."]
            missed_files_str = "<br>".join(missed_files)
            QMessageBox().warning(
                self,
                "Missed Files",
                f"Cannot find files:<br>{missed_files_str}<br>Set proper base directory using <i>Select directory</i>",
            )
        else:
            self.update_files_list(res_list)

    def find_all(self):
        # When it fails again then there is a need to provide test
        # with input that causes crash
        paths = glob(str(self.paths_input.text()))

        paths = sorted(x for x in (set(paths) - self.files_to_proceed) if not os.path.isdir(x))
        if len(paths) > 0:
            self.update_files_list(paths)

        else:
            QMessageBox.warning(self, "No new files", "No new files found", QMessageBox.StandardButton.Ok)

    def update_files_list(self, paths):
        dialog = AcceptFiles(paths)
        if dialog.exec_():
            new_paths = dialog.get_files()
            for path in new_paths:
                self.selected_files.addItem(FileListItem(path))
            self.files_to_proceed.update(new_paths)
            self.file_list_changed.emit(self.files_to_proceed)

    def select_files(self):
        dial = QFileDialog(self, "Select files")
        dial.setDirectory(self.settings.get(IO_BATCH_DIRECTORY, self.settings.get(OPEN_DIRECTORY, str(Path.home()))))
        dial.setFileMode(QFileDialog.FileMode.ExistingFiles)
        if dial.exec_():
            self.settings.set(IO_BATCH_DIRECTORY, os.path.dirname(str(dial.selectedFiles()[0])))
            new_paths = sorted(set(map(str, dial.selectedFiles())) - self.files_to_proceed)
            for path in new_paths:
                self.selected_files.addItem(FileListItem(path))
            self.files_to_proceed.update(new_paths)
            self.file_list_changed.emit(self.files_to_proceed)

    def select_directory(self):
        dial = SelectDirectoryDialog(
            settings=self.settings,
            settings_path=[IO_BATCH_DIRECTORY, OPEN_DIRECTORY],
            default_directory=None,
            parent=self,
        )
        if dial.exec_():
            self.paths_input.setText(dial.selectedFiles()[0])

    def file_chosen(self):
        self.delete_button.setEnabled(True)

    def delete_element(self):
        item = self.selected_files.takeItem(self.selected_files.currentRow())
        self.files_to_proceed.remove(item.file_path)
        self.file_list_changed.emit(self.files_to_proceed)
        if self.selected_files.count() == 0:
            self.delete_button.setDisabled(True)

    def clean(self):
        self.selected_files.clear()
        self.files_to_proceed.clear()
        self.file_list_changed.emit(self.files_to_proceed)

    def get_paths(self):
        return sorted(self.files_to_proceed)



# pylint: disable=no-self-use

import os

import numpy as np
import pytest
from napari.layers import Image, Labels, Layer

from PartSegCore.analysis import ProjectTuple
from PartSegCore.mask.io_functions import LoadROIFromTIFF
from PartSegCore.napari_plugins.load_image import napari_get_reader as napari_get_reader_image
from PartSegCore.napari_plugins.load_mask_project import napari_get_reader as napari_get_reader_mask
from PartSegCore.napari_plugins.load_masked_image import napari_get_reader as napari_get_reader_mask_image
from PartSegCore.napari_plugins.load_roi_project import napari_get_reader as napari_get_reader_roi
from PartSegCore.napari_plugins.loader import project_to_layers
from PartSegCore.napari_plugins.save_mask_roi import napari_write_labels
from PartSegCore.napari_plugins.save_tiff_layer import (
    napari_write_images,
)
from PartSegCore.napari_plugins.save_tiff_layer import (
    napari_write_labels as napari_write_labels_tiff,
)
from PartSegImage import GenericImageReader
from PartSegImage import Image as PImage


def test_project_to_layers_analysis(analysis_segmentation):
    analysis_segmentation.roi_info.alternative["test"] = np.zeros(analysis_segmentation.image.shape, dtype=np.uint8)
    res = project_to_layers(analysis_segmentation)
    assert len(res) == 3
    l1 = Layer.create(*res[0])
    assert isinstance(l1, Image)
    assert l1.name == "channel 1"
    assert np.allclose(l1.scale[1:] / 1e9, analysis_segmentation.image.spacing)
    l2 = Layer.create(*res[1])
    assert isinstance(l2, Labels)
    assert l2.name == "ROI"
    assert np.allclose(l2.scale[1:] / 1e9, analysis_segmentation.image.spacing)
    l3 = Layer.create(*res[2])
    assert isinstance(l3, Labels)
    assert l3.name == "test"
    assert np.allclose(l3.scale[1:] / 1e9, analysis_segmentation.image.spacing)
    assert not l3.visible


def test_project_to_layers_roi():
    data = np.zeros((1, 1, 10, 10, 10), dtype=np.uint8)
    img = PImage(data, image_spacing=(1, 1, 1), name="ROI")
    proj = ProjectTuple(file_path="", image=img)
    res = project_to_layers(proj)
    assert len(res) == 1
    assert isinstance(res[0][0], np.ndarray)
    assert res[0][2] == "labels"


def test_project_to_layers_mask(stack_segmentation1):
    res = project_to_layers(stack_segmentation1)
    assert len(res) == 2
    assert res[0][2] == "image"


@pytest.fixture()
def load_data(data_test_dir):
    def _load_data(file_name, reader_hook):
        file_path = os.path.join(data_test_dir, file_name)
        res = reader_hook(file_path)
        assert res is not None
        return res(file_path)

    return _load_data


@pytest.mark.parametrize(
    "file_name",
    ["test_nucleus.tif", "test_lsm.lsm", "Image0003_01.oif", "test_czi.czi", "N2A_H2BGFP_dapi_falloidin_cycling1.oib"],
)
def test_read_images(load_data, file_name):
    data = load_data(file_name, napari_get_reader_image)
    assert isinstance(data, list)


def test_read_mask_project(load_data):
    data = load_data("test_nucleus.seg", napari_get_reader_mask)
    assert len(data) == 1
    assert data[0][2] == "labels"


def test_read_masked_image(load_data):
    data = load_data("test_nucleus.tif", napari_get_reader_mask_image)
    assert len(data) == 2
    assert data[0][2] == "image"
    assert data[1][2] == "labels"


def test_load_roi_project(load_data):
    data = load_data("stack1_component1.tgz", napari_get_reader_roi)
    assert len(data) == 4
    assert data[0][2] == "image"
    assert data[1][2] == "image"
    assert data[2][2] == "labels"
    assert data[3][2] == "labels"


def test_write_labels(tmp_path):
    data = np.zeros((1, 10, 10, 10), dtype=np.uint8)
    assert napari_write_labels(str(tmp_path / "test.seg"), [], {"scale": [10, 10]}) is None
    assert not (tmp_path / "test.seg").exists()

    assert napari_write_labels(str(tmp_path / "test.seg"), data, {"scale": [10, 10]}) == str(tmp_path / "test.seg")
    assert (tmp_path / "test.seg").exists()

    assert napari_write_labels(str(tmp_path / "test.txt"), data, {"scale": [10, 10]}) is None
    assert not (tmp_path / "test.txt").exists()


def test_save_load_axis_order(tmp_path):
    data = np.zeros((1, 10, 20, 30), dtype=np.uint8)
    layer = Labels(data)
    data_path = str(tmp_path / "test.tif")
    assert napari_write_labels_tiff(data_path, *layer.as_layer_data_tuple()[:2])
    proj = LoadROIFromTIFF.load([data_path])
    assert proj.roi_info.roi.shape == data.shape
    assert napari_write_labels_tiff(str(tmp_path / "test.seg"), *layer.as_layer_data_tuple()[:2]) is None


@pytest.fixture(params=[(1, 4), (1, 3), (3, 4), (3, 3), (10, 4)])
def image_layer_tuples(request):
    res = []
    for i in range(request.param[0]):
        data = np.zeros((1, 10, 20, 30)[-request.param[1] :], dtype=np.uint8)
        data[:, 1:-1, 1:-1] = i + 1
        res.append(Image(data, scale=(1, 1, 1, 1)[-request.param[1] :]).as_layer_data_tuple())
    return res


def test_napari_write_images(image_layer_tuples, tmp_path):
    data_path = str(tmp_path / "test.tif")
    assert len(napari_write_images(data_path, image_layer_tuples)) == 1
    image = GenericImageReader.read_image(data_path)
    assert image.channels == len(image_layer_tuples)


def test_write_multichannel(tmp_path):
    data = np.zeros((1, 10, 20, 30, 5), dtype=np.uint8)
    data[:, 1:-1, 1:-1, 1:-1] = 1
    layer = Image(data)
    data_path = str(tmp_path / "test.tif")
    assert napari_write_images(data_path, [layer.as_layer_data_tuple()])
    image = GenericImageReader.read_image(data_path)
    assert image.channels == 5


def test_different_shapes(tmp_path):
    layer1 = Image(np.eye(10, dtype=np.uint8))
    layer2 = Image(np.eye(20, dtype=np.uint8))
    data_path = str(tmp_path / "test.tif")
    assert not napari_write_images(data_path, [layer1.as_layer_data_tuple(), layer2.as_layer_data_tuple()])
    assert not (tmp_path / "test.tif").exists()



import typing
import warnings
from copy import deepcopy

from qtpy.QtCore import Signal

from PartSeg.common_backend.base_settings import BaseSettings, SaveSettingsDescription
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis.analysis_utils import SegmentationPipeline
from PartSegCore.analysis.calculation_plan import CalculationPlan
from PartSegCore.analysis.io_utils import MaskInfo, ProjectTuple
from PartSegCore.analysis.load_functions import load_metadata
from PartSegCore.analysis.measurement_calculation import MeasurementProfile
from PartSegCore.io_utils import PointsInfo
from PartSegCore.json_hooks import PartSegEncoder
from PartSegCore.project_info import HistoryElement
from PartSegCore.roi_info import ROIInfo
from PartSegCore.utils import EventedDict, ProfileDict


class PartSettings(BaseSettings):
    """
    last_executed_algorithm - parameter for caring last used algorithm
    """

    compare_segmentation_change = Signal(ROIInfo)
    roi_profiles_changed = Signal()
    roi_pipelines_changed = Signal()
    measurement_profiles_changed = Signal()
    batch_plans_changed = Signal()
    json_encoder_class = PartSegEncoder
    load_metadata = staticmethod(load_metadata)
    last_executed_algorithm: str
    save_locations_keys: typing.ClassVar[typing.List[str]] = [
        "open_directory",
        "save_directory",
        "export_directory",
        "batch_plan_directory",
        "multiple_open_directory",
    ]

    def __init__(self, json_path, profile_name="default"):
        super().__init__(json_path, profile_name)
        self._mask = None
        self.compare_segmentation = None
        self._segmentation_pipelines_dict = ProfileDict(klass={"*": {"*": SegmentationPipeline}})
        self._segmentation_profiles_dict = ProfileDict(klass={"*": {"*": ROIExtractionProfile}})
        self._batch_plans_dict = ProfileDict(klass={"*": {"*": CalculationPlan}})
        self._measurement_profiles_dict = ProfileDict(klass={"*": {"*": MeasurementProfile}})
        self._segmentation_profiles_dict.connect("", self.roi_profiles_changed.emit, maxargs=0)
        self._segmentation_pipelines_dict.connect("", self.roi_pipelines_changed.emit, maxargs=0)
        self._measurement_profiles_dict.connect("", self.measurement_profiles_changed.emit, maxargs=0)
        self._batch_plans_dict.connect("", self.batch_plans_changed.emit, maxargs=0)

    def fix_history(self, algorithm_name, algorithm_values):
        """
        set new algorithm parameters to

        :param str algorithm_name:
        :param dict algorithm_values:
        """
        self.history[self.history_index + 1] = self.history[self.history_index + 1].copy(
            update={"roi_extraction_parameters": {"algorithm_name": algorithm_name, "values": algorithm_values}}
        )

    @staticmethod
    def cmp_history_element(el1: HistoryElement, el2: HistoryElement):
        return el1.mask_property == el2.mask_property

    def set_segmentation_to_compare(self, segmentation: ROIInfo):
        self.compare_segmentation = segmentation
        self.compare_segmentation_change.emit(segmentation)

    def _image_changed(self):
        super()._image_changed()
        self._mask = None

    def get_project_info(self) -> ProjectTuple:
        if algorithm_name := self.last_executed_algorithm:
            value = self.get_algorithm(f"algorithms.{algorithm_name}")
            if isinstance(value, EventedDict):
                value = value.as_dict_deep()
            algorithm_val = {
                "algorithm_name": algorithm_name,
                "values": deepcopy(value),
            }
        else:
            algorithm_val = {}
        return ProjectTuple(
            file_path=self.image.file_path,
            image=self.image.substitute(),
            roi_info=self.roi_info,
            additional_layers=self.additional_layers,
            mask=self.mask,
            history=self.history[: self.history_index + 1],
            algorithm_parameters=algorithm_val,
            points=self.points,
        )

    def set_project_info(self, data: typing.Union[ProjectTuple, MaskInfo, PointsInfo]):
        if isinstance(data, MaskInfo):
            self.mask = data.mask_array
            return
        if isinstance(data, PointsInfo):
            self.points = data.points
            return
        if not isinstance(data, ProjectTuple):
            return
        if self.image.file_path == data.image.file_path and self.image.shape == data.image.shape:
            if data.roi_info.roi is not None:
                try:
                    self.image.fit_array_to_image(data.roi_info.roi)
                    self.mask = data.mask
                except ValueError:
                    self.image = data.image.substitute()
            else:
                self.mask = data.mask
        else:
            self.image = data.image.substitute(mask=data.mask)
        self.roi = data.roi_info
        self._additional_layers = data.additional_layers
        self.additional_layers_changed.emit()
        self.set_history(data.history[:])
        if data.algorithm_parameters:
            self.last_executed_algorithm = data.algorithm_parameters["algorithm_name"]
            self.set_algorithm(
                f"algorithms.{self.last_executed_algorithm}", deepcopy(data.algorithm_parameters["values"])
            )
            self.algorithm_changed.emit()

    def get_save_list(self) -> typing.List[SaveSettingsDescription]:
        return [
            *super().get_save_list(),
            SaveSettingsDescription("segmentation_pipeline_save.json", self._segmentation_pipelines_dict),
            SaveSettingsDescription("segmentation_profiles_save.json", self._segmentation_profiles_dict),
            SaveSettingsDescription("statistic_profiles_save.json", self._measurement_profiles_dict),
            SaveSettingsDescription("batch_plans_save.json", self._batch_plans_dict),
        ]

    @property
    def segmentation_pipelines(self) -> typing.Dict[str, SegmentationPipeline]:
        warnings.warn("segmentation_pipelines is deprecated, use roi_pipelines", DeprecationWarning, stacklevel=2)
        return self.roi_pipelines

    @property
    def roi_pipelines(self) -> typing.Dict[str, SegmentationPipeline]:
        return self._segmentation_pipelines_dict.get(self._current_roi_dict, EventedDict())

    @property
    def segmentation_profiles(self) -> typing.Dict[str, ROIExtractionProfile]:
        warnings.warn("segmentation_profiles is deprecated, use roi_profiles", DeprecationWarning, stacklevel=2)
        return self.roi_profiles

    @property
    def roi_profiles(self) -> typing.Dict[str, ROIExtractionProfile]:
        return self._segmentation_profiles_dict.get(self._current_roi_dict, EventedDict())

    @property
    def batch_plans(self) -> typing.Dict[str, CalculationPlan]:
        return self._batch_plans_dict.get(self._current_roi_dict, EventedDict())

    @property
    def measurement_profiles(self) -> typing.Dict[str, MeasurementProfile]:
        return self._measurement_profiles_dict.get(self._current_roi_dict, EventedDict())



"""
This module contains non gui Qt based components
"""

import os.path
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from napari.settings import NapariSettings


def napari_get_settings(path=None) -> "NapariSettings":
    from napari.settings import get_settings as _napari_get_settings

    if path is not None:
        path = os.path.join(path, "settings.yaml")

    try:
        return _napari_get_settings(path)
    except Exception:  # pylint: disable=broad-except
        return _napari_get_settings()


__all__ = ("napari_get_settings",)



import inspect
import math
import textwrap
import typing
import warnings
from abc import ABC, ABCMeta, abstractmethod
from functools import wraps
from importlib.metadata import version

from local_migrator import REGISTER, class_to_str
from pydantic import BaseModel as PydanticBaseModel
from pydantic import create_model, validator
from pydantic.fields import FieldInfo
from typing_extensions import Annotated

from PartSegCore.utils import BaseModel
from PartSegImage import Channel

if typing.TYPE_CHECKING:
    from pydantic.fields import ModelField

try:
    # pydantic 1
    from pydantic.fields import UndefinedType
    from pydantic.main import ModelMetaclass

    def field_serializer(_):
        def decorator(func):
            return func

        return decorator

except ImportError:
    # pydantic 2
    from pydantic import field_serializer
    from pydantic._internal._model_construction import ModelMetaclass
    from pydantic_core import PydanticUndefinedType as UndefinedType


class AlgorithmDescribeNotFound(Exception):
    """
    When algorithm description not found
    """


class AlgorithmProperty:
    """
    This class is used to verbose describe algorithm parameters

    :param str name: name of parameter used in code
    :param str user_name: name presented to user in interface
    :param default_value: initial value which be used during interface generation
    :param str help_text: toll tip presented to user when keep mouse over widget
    :type value_type: type
    """

    def __init__(
        self,
        name: str,
        user_name: str,
        default_value: typing.Union[str, float, object],
        options_range=None,
        possible_values=None,
        value_type=None,
        help_text="",
        per_dimension=False,
        mgi_options=None,
        **kwargs,
    ):
        if "property_type" in kwargs:
            warnings.warn("property_type is deprecated, use value_type instead", DeprecationWarning, stacklevel=2)
            value_type = kwargs["property_type"]
            del kwargs["property_type"]
        if kwargs:
            raise ValueError(", ".join(kwargs.keys()) + " are not expected")

        self.name = name
        self.user_name = user_name
        if isinstance(possible_values, list):
            self.value_type = list
        elif value_type is not None:
            self.value_type = value_type
        else:
            self.value_type = type(default_value)
        self.default_value = default_value
        self.range = options_range
        self.possible_values = possible_values
        self.help_text = help_text
        self.per_dimension = per_dimension
        self.mgi_options = mgi_options if mgi_options is not None else {}
        if self.value_type is list and default_value not in possible_values:
            raise ValueError(f"default_value ({default_value}) should be one of possible values ({possible_values}).")

    def __repr__(self):
        return (
            f"{self.__class__.__module__}.{self.__class__.__name__}(name='{self.name}',"
            f" user_name='{self.user_name}', "
            f"default_value={self.default_value}, type={self.value_type}, range={self.range},"
            f"possible_values={self.possible_values})"
        )


class _GetDescriptionClass:
    __slots__ = ("_name",)

    def __init__(self):
        self._name = None

    def __set_name__(self, owner, name: str):
        if self._name is None:
            self._name = name

    def __get__(self, obj, klass):
        if klass is None:
            klass = type(obj)

        name = typing.cast(str, self._name)
        fields_dkt = {
            field.name: (
                Annotated[field.value_type, field.user_name, field.range, field.help_text],
                field.default_value,
            )
            for field in klass.get_fields()
            if not isinstance(field, str)
        }

        model = create_model(name, **fields_dkt)
        model.__qualname__ = f"{klass.__qualname__}.{name}"
        setattr(klass, name, model)
        return model


def _partial_abstractmethod(funcobj):
    funcobj.__is_partial_abstractmethod__ = True
    return funcobj


class AlgorithmDescribeBaseMeta(ABCMeta):
    def __new__(cls, name, bases, attrs, **kwargs):
        cls2 = super().__new__(cls, name, bases, attrs, **kwargs)
        if (
            not inspect.isabstract(cls2)
            and hasattr(cls2.get_fields, "__is_partial_abstractmethod__")
            and cls2.__argument_class__ is None
        ):
            raise RuntimeError("class need to have __argument_class__ set or get_fields functions defined")
        cls2.__new_style__ = getattr(cls2.get_fields, "__is_partial_abstractmethod__", False)
        return cls2


class AlgorithmDescribeBase(ABC, metaclass=AlgorithmDescribeBaseMeta):
    """
    This is abstract class for all algorithm exported to user interface.
    Based on get_name and get_fields methods the interface will be generated
    For each group of algorithm base abstract class will add additional methods
    """

    __argument_class__: typing.Optional[typing.Type[PydanticBaseModel]] = None
    __new_style__: bool

    @classmethod
    def get_doc_from_fields(cls):
        resp = "{\n"
        for el in get_fields_from_algorithm(cls):
            if isinstance(el, AlgorithmProperty):
                resp += f"  {el.name}: {el.value_type} - "
                if el.help_text:
                    resp += el.help_text
                resp += f"(default values: {el.default_value})\n"
        resp += "}\n"
        return resp

    @classmethod
    @abstractmethod
    def get_name(cls) -> str:
        """
        Algorithm name. It will be used during interface generating and in registering
        to proper :py:class:`PartSeg.PartSegCore.algorithm_describe_base.Register`.

        :return: name of algorithm
        """
        raise NotImplementedError

    @classmethod
    @_partial_abstractmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        """
        This function return list of parameters needed by algorithm. It is used for generate form in User Interface

        :return: list of algorithm parameters and comments
        """
        if hasattr(cls, "__argument_class__") and cls.__argument_class__ is not None:
            warnings.warn(
                "Class has __argument_class__ defined, one should not use get_fields",
                category=FutureWarning,
                stacklevel=2,
            )
            return base_model_to_algorithm_property(cls.__argument_class__)
        raise NotImplementedError

    @classmethod
    def _get_fields(cls):
        return base_model_to_algorithm_property(cls.__argument_class__) if cls.__new_style__ else cls.get_fields()

    @classmethod
    def get_fields_dict(cls) -> typing.Dict[str, AlgorithmProperty]:
        return {v.name: v for v in get_fields_from_algorithm(cls) if isinstance(v, AlgorithmProperty)}

    @classmethod
    def get_default_values(cls):
        if cls.__new_style__:
            return cls.__argument_class__()  # pylint: disable=not-callable
        return {
            el.name: (
                {
                    "name": el.default_value,
                    "values": el.possible_values[el.default_value].get_default_values(),
                }
                if issubclass(el.value_type, AlgorithmDescribeBase)
                else el.default_value
            )
            for el in cls.get_fields()
            if isinstance(el, AlgorithmProperty)
        }


def get_fields_from_algorithm(ald_desc: AlgorithmDescribeBase) -> typing.List[typing.Union[AlgorithmProperty, str]]:
    if ald_desc.__new_style__:
        return base_model_to_algorithm_property(ald_desc.__argument_class__)
    return ald_desc.get_fields()


def is_static(fun):
    if fun is None:
        return False
    args = inspect.getfullargspec(fun).args
    return True if len(args) == 0 else args[0] != "self"


AlgorithmType = typing.TypeVar("AlgorithmType", bound=typing.Type[AlgorithmDescribeBase])


class Register(typing.Dict, typing.Generic[AlgorithmType]):
    """
    Dict used for register :class:`.AlgorithmDescribeBase` classes.
    All registers from `PartSeg.PartSegCore.register` are this
    :param class_methods: list of method which should be implemented as class method it will be checked during add
    as args or with :meth:`.Register.register`  method
    :param methods: list of method which should be instance method
    """

    def __init__(self, *args: AlgorithmType, class_methods=None, methods=None, suggested_base_class=None, **kwargs):
        """
        :param class_methods: list of method which should be class method
        :param methods: list of method which should be instance method
        :param kwargs: elements passed to OrderedDict constructor (may be initial elements). I suggest to not use this.
        """
        super().__init__(**kwargs)
        self.suggested_base_class = suggested_base_class
        self.class_methods = (
            list(class_methods) if class_methods else getattr(suggested_base_class, "need_class_method", [])
        )
        self.methods = list(methods) if methods else getattr(suggested_base_class, "need_method", [])
        self._old_mapping = {}
        for el in args:
            self.register(el)

    def values(self) -> typing.Iterable[AlgorithmType]:
        # noinspection PyTypeChecker
        return typing.cast(typing.Iterable[AlgorithmType], super().values())

    def __eq__(self, other):
        return (
            super().__eq__(other)
            and isinstance(other, Register)
            and self.class_methods == other.class_methods
            and self.methods == other.methods
            and self.suggested_base_class == other.suggested_base_class
        )

    def __ne__(self, other):
        return not self.__eq__(other)

    def __getitem__(self, item) -> AlgorithmType:
        # FIXME add better strategy to get proper class when there is conflict of names
        try:
            return typing.cast(AlgorithmType, super().__getitem__(item))
        except KeyError:
            return typing.cast(AlgorithmType, super().__getitem__(self._old_mapping[item]))

    def __contains__(self, item):
        return super().__contains__(item) or item in self._old_mapping

    def register(
        self, value: AlgorithmType, replace: bool = False, old_names: typing.Optional[typing.List[str]] = None
    ):
        """
        Function for registering :class:`.AlgorithmDescribeBase` based algorithms
        :param value: algorithm to register
        :param replace: replace existing algorithm, be patient with
        :param old_names: list of old names for registered class
        """
        self.check_function(value, "get_name", True)
        try:
            name = value.get_name()
        except (NotImplementedError, AttributeError):
            raise ValueError(f"Class {value} need to implement classmethod 'get_name'") from None
        if name in self and not replace:
            raise ValueError(
                f"Object {self[name]} with this name: '{name}' already exist and register is not in replace mode"
            )
        if not isinstance(name, str):
            raise ValueError(f"Function get_name of class {value} need return string not {type(name)}")
        self[name] = value
        if old_names is not None:
            # FIXME add better strategy to get proper class when there is conflict of names
            for old_name in old_names:
                if old_name in self._old_mapping and not replace:
                    raise ValueError(
                        f"Old value mapping for name '{old_name}' already registered."
                        f" Currently pointing to {self._old_mapping[old_name]}"
                    )
                self._old_mapping[old_name] = name
        return value

    @staticmethod
    def check_function(ob, function_name, is_class):
        fun = getattr(ob, function_name, None)
        if not is_class and not inspect.isfunction(fun):
            raise ValueError(f"Class {ob} need to define method {function_name}")
        if is_class and not inspect.ismethod(fun) and not is_static(fun):
            raise ValueError(f"Class {ob} need to define classmethod '{function_name}'")

    def __setitem__(self, key: str, value: AlgorithmType):
        if not issubclass(value, AlgorithmDescribeBase):
            raise ValueError(
                f"Class {value} need to be subclass of {AlgorithmDescribeBase.__module__}.AlgorithmDescribeBase"
            )
        self.check_function(value, "get_name", True)
        self.check_function(value, "get_fields", True)
        try:
            val = value.get_name()
        except (NotImplementedError, AttributeError):
            raise ValueError(f"Class {value} need to implement classmethod 'get_name'") from None
        if not isinstance(val, str):
            raise ValueError(f"Function get_name of class {value} need return string not {type(val)}")
        if key != val:
            raise ValueError("Object need to be registered under name returned by get_name function")
        if not value.__new_style__:
            try:
                val = value.get_fields()
                if not isinstance(val, list):
                    raise ValueError(f"Function get_fields of class {value} need return list not {type(val)}")
            except NotImplementedError:
                raise ValueError(f"Method get_fields of class {value} need to be implemented") from None
        for el in self.class_methods:
            self.check_function(value, el, True)
        for el in self.methods:
            self.check_function(value, el, False)

        super().__setitem__(key, value)

    def get_default(self) -> str:
        """
        Calculate default algorithm name for given dict.

        :return: name of algorithm
        """
        try:
            return next(iter(self.keys()))
        except StopIteration:
            raise ValueError("Register does not contain any algorithm.") from None


class AddRegisterMeta(ModelMetaclass):
    def __new__(cls, name, bases, attrs, **kwargs):
        methods = kwargs.pop("methods", [])
        suggested_base_class = kwargs.pop("suggested_base_class", None)
        class_methods = kwargs.pop("class_methods", [])
        cls2 = super().__new__(cls, name, bases, attrs, **kwargs)  # pylint: disable=too-many-function-args
        cls2.__register__ = Register(
            class_methods=class_methods, methods=methods, suggested_base_class=suggested_base_class
        )
        return cls2

    def __getitem__(self, item) -> AlgorithmType:
        return self.__register__[item]

    def __contains__(self, item) -> bool:
        return self.__register__.__contains__(item)

    def get(self, item, default=None):
        return self.__register__.get(item, default)


class AlgorithmSelection(BaseModel, metaclass=AddRegisterMeta):  # pylint: disable=E1139
    """
    Base class for algorithm selection.
    For given algorithm there should be Register instance set __register__ class variable.
    """

    name: str
    values: typing.Union[typing.Dict[str, typing.Any], PydanticBaseModel]
    class_path: str = ""
    if typing.TYPE_CHECKING:
        __register__: Register

    class Config:
        smart_union = True

    @validator("name")
    def check_name(cls, v):
        if v not in cls.__register__:
            raise ValueError(f"Missed algorithm {v}")
        return v

    @validator("class_path", always=True)
    def update_class_path(cls, v, values):
        if v or "name" not in values:
            return v
        klass = cls.__register__[values["name"]]
        return class_to_str(klass)

    @field_serializer("values")
    def val_serializer(self, value, _info):
        if isinstance(value, PydanticBaseModel):
            return value.dict()
        return value

    @validator("values", pre=True)
    def update_values(cls, v, values):
        # FIXME add better strategy to get proper class when there is conflict of names
        if "name" not in values or not isinstance(v, dict):
            return v
        klass = cls.__register__[values["name"]]
        if not klass.__new_style__ or not klass.__argument_class__.__fields__:
            return v

        dkt_migrated = REGISTER.migrate_data(class_to_str(klass.__argument_class__), {}, v)
        return klass.__argument_class__(**dkt_migrated)

    @classmethod
    def register(
        cls, value: AlgorithmType, replace=False, old_names: typing.Optional[typing.List[str]] = None
    ) -> AlgorithmType:
        """
        Function for registering :class:`.AlgorithmDescribeBase` based algorithms
        :param value: algorithm to register
        :param replace: replace existing algorithm, be patient with
        :param old_names: list of old names for registered class
        """
        return cls.__register__.register(value, replace=replace, old_names=old_names)

    @classmethod
    def get_default(cls):
        name = cls.__register__.get_default()
        return cls(name=name, values=cls[name].get_default_values())

    def algorithm(self):
        return self.__register__[self.name]


class ROIExtractionProfileMeta(ModelMetaclass):
    def __new__(cls, name, bases, attrs, **kwargs):
        cls2 = super().__new__(cls, name, bases, attrs, **kwargs)  # pylint: disable=too-many-function-args

        def allow_positional_args(func):
            @wraps(func)
            def _wraps(self, *args, **kwargs):
                if args:
                    warnings.warn(
                        "Positional arguments are deprecated, use keyword arguments instead",
                        FutureWarning,
                        stacklevel=2,
                    )
                    kwargs.update(dict(zip(self.__fields__, args)))
                return func(self, **kwargs)

            return _wraps

        cls2.__init__ = allow_positional_args(cls2.__init__)
        return cls2


class ROIExtractionProfile(BaseModel, metaclass=ROIExtractionProfileMeta):  # pylint: disable=E1139
    """
    :ivar str ~.name: name for segmentation profile
    :ivar str ~.algorithm: Name of algorithm
    :ivar dict ~.values: algorithm parameters
    """

    name: str
    algorithm: str
    values: typing.Any

    @validator("values")
    def validate_values(cls, v, values):  # pylint: disable=no-self-use
        if not isinstance(v, dict):
            return v
        if "algorithm" not in values:
            return v
        from PartSegCore.analysis import AnalysisAlgorithmSelection
        from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection

        name = values["algorithm"]
        is_analysis = name in AnalysisAlgorithmSelection
        is_mask = name in MaskAlgorithmSelection
        if is_analysis == is_mask:
            return v
        algorithm = AnalysisAlgorithmSelection[name] if is_analysis else MaskAlgorithmSelection[name]
        if not algorithm.__new_style__:
            return v
        dkt_migrated = REGISTER.migrate_data(class_to_str(algorithm.__argument_class__), {}, v)
        return algorithm.__argument_class__(**dkt_migrated)

    def pretty_print(self, algorithm_dict):
        if isinstance(algorithm_dict, AlgorithmSelection):
            algorithm_dict = algorithm_dict.__register__
        try:
            algorithm = algorithm_dict[self.algorithm]
        except KeyError:
            return f"{self}\n "
        values = self.values if isinstance(self.values, dict) else self.values.dict()
        if self.name in {"", "Unknown"}:
            return (
                "ROI extraction profile\nAlgorithm: "
                + self.algorithm
                + "\n"
                + self._pretty_print(values, algorithm.get_fields_dict())
            )
        return (
            f"ROI extraction profile name: {self.name}\nAlgorithm: {self.algorithm}\n"
            f"{self._pretty_print(values, algorithm.get_fields_dict())}"
        )

    @classmethod
    def _pretty_print(
        cls, values: typing.MutableMapping, translate_dict: typing.Dict[str, AlgorithmProperty], indent=0
    ):
        if not isinstance(values, typing.MutableMapping):
            return textwrap.indent(str(values), " " * indent)
        res = ""
        for k, v in values.items():
            if k not in translate_dict:
                if isinstance(v, typing.MutableMapping):
                    res += " " * indent + f"{k}: {cls._pretty_print(v, {}, indent + 2)}\n"
                else:
                    res += " " * indent + f"{k}: {v}\n"
                continue
            desc = translate_dict[k]
            res += " " * indent + desc.user_name + ": "
            if issubclass(desc.value_type, Channel) and not isinstance(v, Channel):
                res += str(Channel(v))
            elif issubclass(desc.value_type, AlgorithmDescribeBase):
                if isinstance(v, AlgorithmSelection):
                    name = v.name
                    values_ = v.values
                else:
                    name = v["name"]
                    values_ = v["values"]
                res += desc.possible_values[name].get_name()
                if values_:
                    res += "\n"
                    res += cls._pretty_print(values_, desc.possible_values[name].get_fields_dict(), indent + 2)
            elif isinstance(v, typing.MutableMapping):
                res += cls._pretty_print(v, {}, indent + 2)
            else:
                res += str(v)
            res += "\n"
        return res[:-1]

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__)
            and self.name == other.name
            and self.algorithm == other.algorithm
            and self.values == other.values
        )


def _next_after(type_, value, inf):
    if issubclass(type_, int):
        if inf == math.inf:
            return value + 1
        return value - 1
    if hasattr(math, "nextafter"):
        # TODO fix after drop python 3.8
        return math.nextafter(value, inf)
    return value


def _next_after_with_none(type_, value, inf):
    if value is None:
        return None
    return _next_after(type_, value, inf)


def _calc_value_range(field_info: FieldInfo):
    if field_info.metadata is None:
        return (0, 1000)

    import annotated_types as at

    value_range = (0, 1000)
    for el in field_info.metadata:
        if isinstance(el, at.Ge):
            value_range = el.ge, value_range[1]
        elif isinstance(el, at.Gt):
            value_range = _next_after(field_info.annotation, el.gt, math.inf), value_range[1]
        elif isinstance(el, at.Le):
            value_range = value_range[0], el.le
        elif isinstance(el, at.Lt):
            value_range = value_range[0], _next_after(field_info.annotation, el.lt, -math.inf)
    return value_range


def _field_to_algorithm_property_pydantic_2(name: str, field_info: FieldInfo):
    user_name = field_info.title

    value_range = None
    possible_values = None

    value_type = field_info.annotation
    default_value = field_info.default
    help_text = field_info.description
    if user_name is None:
        user_name = name.replace("_", " ").capitalize()
    if not hasattr(value_type, "__origin__"):
        if issubclass(value_type, (int, float)):
            value_range = _calc_value_range(field_info)

        if issubclass(field_info.annotation, AlgorithmSelection):
            value_type = AlgorithmDescribeBase
            if isinstance(field_info.default, UndefinedType):
                default_value = field_info.default_factory().name
            else:
                default_value = field_info.default.name
            possible_values = field_info.annotation.__register__

    extra = field_info.json_schema_extra or {}

    return AlgorithmProperty(
        name=name,
        user_name=user_name,
        default_value=default_value,
        options_range=value_range,
        value_type=value_type,
        possible_values=possible_values,
        help_text=help_text,
        mgi_options=extra.get("options", {}),
    )


def _field_to_algorithm_property_pydantic_1(name: str, field: "ModelField"):
    user_name = field.field_info.title
    value_range = None
    possible_values = None

    value_type = getattr(field, "annotation", field.type_)
    default_value = field.field_info.default
    help_text = field.field_info.description
    if user_name is None:
        user_name = name.replace("_", " ").capitalize()
    if not hasattr(field.type_, "__origin__"):
        if issubclass(field.type_, (int, float)):
            value_range = (
                field.field_info.ge or _next_after_with_none(field.type_, field.field_info.gt, math.inf) or 0,
                field.field_info.le or _next_after_with_none(field.type_, field.field_info.lt, -math.inf) or 1000,
            )
        if issubclass(field.type_, AlgorithmSelection):
            value_type = AlgorithmDescribeBase
            if isinstance(field.field_info.default, UndefinedType):
                default_value = field.field_info.default_factory().name
            else:
                default_value = field.field_info.default.name
            possible_values = field.type_.__register__

    return AlgorithmProperty(
        name=name,
        user_name=user_name,
        default_value=default_value,
        options_range=value_range,
        value_type=value_type,
        possible_values=possible_values,
        help_text=help_text,
        mgi_options=field.field_info.extra.get("options", {}),
    )


def base_model_to_algorithm_property_pydantic_1(
    obj: typing.Type[BaseModel],
) -> typing.List[typing.Union[str, AlgorithmProperty]]:
    """
    Convert pydantic model to list of AlgorithmPropert nad strings.

    :param obj:
    :return:
    """
    res = []
    value: ModelField
    if hasattr(obj, "header") and obj.header():
        res.append(obj.header())
    for name, value in obj.__fields__.items():
        ap = _field_to_algorithm_property(name, value)
        if value.field_info.extra.get("hidden", False):
            continue
        pos = len(res)
        if "position" in value.field_info.extra:
            pos = value.field_info.extra["position"]
        if "prefix" in value.field_info.extra:
            res.insert(pos, value.field_info.extra["prefix"])
            pos += 1

        res.insert(pos, ap)

        if "suffix" in value.field_info.extra:
            res.insert(pos + 1, value.field_info.extra["suffix"])
    return res


def base_model_to_algorithm_property_pydantic_2(
    obj: typing.Type[BaseModel],
) -> typing.List[typing.Union[str, AlgorithmProperty]]:
    """
    Convert pydantic model to list of AlgorithmPropert nad strings.

    :param obj:
    :return:
    """
    res = []
    field_info: FieldInfo
    if hasattr(obj, "header") and obj.header():
        res.append(obj.header())
    for name, field_info in obj.__fields__.items():
        ap = _field_to_algorithm_property(name, field_info)
        extra = field_info.json_schema_extra or {}
        if extra.get("hidden", False):
            continue
        pos = len(res)
        if "position" in extra:
            pos = extra["position"]
        if "prefix" in extra:
            res.insert(pos, extra["prefix"])
            pos += 1

        res.insert(pos, ap)

        if "suffix" in extra:
            res.insert(pos + 1, extra["suffix"])
    return res


if version("pydantic") < "2":
    _field_to_algorithm_property = _field_to_algorithm_property_pydantic_1
    base_model_to_algorithm_property = base_model_to_algorithm_property_pydantic_1
else:
    _field_to_algorithm_property = _field_to_algorithm_property_pydantic_2
    base_model_to_algorithm_property = base_model_to_algorithm_property_pydantic_2



from qtpy.QtCore import QThread
from qtpy.QtWidgets import QDialog, QHBoxLayout, QLabel, QProgressBar, QPushButton

from PartSeg.common_backend.progress_thread import ExecuteFunctionThread, ProgressTread

try:
    from qtpy import QT5
except ImportError:  # pragma: no cover
    QT5 = True


class WaitingDialog(QDialog):
    def __init__(self, thread: QThread, text="", parent=None, exception_hook=None):
        super().__init__(parent)
        label = QLabel(text)
        self.exception_hook = exception_hook
        self.progress = QProgressBar()
        self.progress.setRange(0, 0)
        self.cancel_btn = QPushButton("Cancel")
        self.cancel_btn.setDisabled(True)
        self._error_caught = False
        layout = QHBoxLayout()
        layout.addWidget(label)
        layout.addWidget(self.progress)
        layout.addWidget(self.cancel_btn)
        self.cancel_btn.clicked.connect(self.close)
        thread.finished.connect(self.accept_if_no_reject)
        self.thread_to_wait = thread
        self.setLayout(layout)
        self.setResult(QDialog.DialogCode.Accepted)
        if isinstance(thread, ProgressTread):
            thread.range_changed.connect(self.progress.setRange)
            thread.step_changed.connect(self.progress.setValue)
            thread.error_signal.connect(self.error_catch)

    def error_catch(self, error):
        self._error_caught = True
        self.reject()
        if self.exception_hook:
            self.exception_hook(error)
        else:
            raise error

    def accept_if_no_reject(self):
        if self._error_caught:
            self.reject()
        else:
            self.accept()

    if QT5:

        def exec(self):
            self.exec_()

    def exec_(self):
        self.thread_to_wait.start()
        return super().exec_()


class ExecuteFunctionDialog(WaitingDialog):
    thread_to_wait: ExecuteFunctionThread

    def __init__(self, fun, args=None, kwargs=None, text="", parent=None, exception_hook=None):
        thread = ExecuteFunctionThread(fun, args, kwargs)
        super().__init__(thread, text=text, parent=parent, exception_hook=exception_hook)

    def get_result(self):
        return self.thread_to_wait.result

    def __repr__(self):
        return (
            f"ExecuteFunctionDialog({self.thread_to_wait.function.__name__}, {self.thread_to_wait.args}, "
            f"{self.thread_to_wait.kwargs}, {self.exception_hook})"
        )



changelog = """
# Changelog

## 0.15.3 - 2024-07-03

### 🚀 Features

- Pydantic 2 compatybility (#1084)

### 🐛 Bug Fixes

- Fix rendering icons in colormap preview (#1040)
- Fix test for validation length of message for sentry-sdk 2.0 release (#1098)
- When fix reader check lowercase extension for validate compatybility (#1097)
- Fix napari 0.5.0 compatybility (#1116)

### 🚜 Refactor

- Fix Qt flags (#1041)
- Fix qt flags in roi mask code (#1042)
- Fix qt flags in roi analysis (#1043)
- Migrate from setup.cfg to `pyproject.toml` (#1070)

### 📚 Documentation

- Allow to use newer release of build docs dependencies (#1057)

### 🧪 Testing

- \[Automatic\] Constraints upgrades: `imagecodecs`, `imageio`, `ipykernel`, `ipython`, `numpy`, `oiffile`, `pandas`, `psygnal`, `pyinstaller`, `qtconsole`, `qtpy`, `sentry-sdk`, `simpleitk`, `superqt`, `tifffile`, `xlsxwriter` (#1020)
- \[Automatic\] Constraints upgrades: `h5py`, `imageio`, `ipython`, `numpy`, `packaging`, `pydantic`, `pyinstaller`, `pyqt5`, `scipy`, `sentry-sdk`, `superqt`, `tifffile`, `xlsxwriter` (#1027)
- \[Automatic\] Constraints upgrades: `imageio`, `magicgui`, `xlsxwriter` (#1030)
- \[Automatic\] Constraints upgrades: `ipykernel`, `pandas`, `qtpy` (#1032)
- \[Automatic\] Constraints upgrades: `imageio`, `ipykernel`, `ipython`, `numpy`, `pandas`, `psygnal`, `pygments`, `pyinstaller`, `qtconsole`, `scipy`, `sentry-sdk`, `simpleitk` (#1035)
- \[Automatic\] Constraints upgrades: `imagecodecs`, `imageio`, `ipykernel`, `magicgui`, `pandas`, `pyinstaller`, `qtawesome`, `sentry-sdk`, `tifffile` (#1048)
- \[Automatic\] Constraints upgrades: `ipykernel`, `numpy`, `pandas`, `partsegcore-compiled-backend`, `pydantic`, `scipy`, `sentry-sdk` (#1058)
- Improve test of PartSegImage (#1072)
- Improve test suite for `PartSegCore` (#1077)
- \[Automatic\] Constraints upgrades: `imageio`, `ipykernel`, `local-migrator`, `napari`, `numpy`, `pandas`, `partsegcore-compiled-backend`, `pyinstaller`, `sentry-sdk`, `tifffile`, `vispy`, `xlsxwriter` (#1063)
- \[Automatic\] Constraints upgrades: `magicgui`, `packaging`, `psygnal`, `pyinstaller`, `sentry-sdk`, `superqt` (#1086)
- \[Automatic\] Constraints upgrades: `psygnal`, `pydantic`, `sentry-sdk`, `vispy` (#1090)
- \[Automatic\] Constraints upgrades: `h5py`, `ipykernel`, `mahotas`, `pandas`, `psygnal`, `pydantic`, `pyinstaller`, `qtawesome`, `scipy`, `sentry-sdk`, `superqt` (#1092)
- \[Automatic\] Constraints upgrades: `imageio`, `tifffile` (#1100)
- \[Automatic\] Constraints upgrades: `pydantic`, `sentry-sdk`, `superqt`, `tifffile` (#1102)
- \[Automatic\] Constraints upgrades: `psygnal`, `pygments`, `qtconsole`, `sentry-sdk`, `superqt`, `tifffile` (#1105)
- \[Automatic\] Constraints upgrades: `imagecodecs`, `magicgui`, `oiffile`, `openpyxl`, `packaging`, `pydantic`, `pyinstaller`, `requests`, `scipy`, `sentry-sdk`, `superqt`, `sympy`, `tifffile`, `vispy` (#1107)

### ⚙️ Miscellaneous Tasks

- \[pre-commit.ci\] pre-commit autoupdate (#1019)
- Remove plugin page preview as it is no longer maintained (#1021)
- \[pre-commit.ci\] pre-commit autoupdate (#1022)
- \[pre-commit.ci\] pre-commit autoupdate (#1026)
- \[pre-commit.ci\] pre-commit autoupdate (#1031)
- \[pre-commit.ci\] pre-commit autoupdate (#1034)
- Use new semgrep configuration (#1039)
- Upload raw coverage information (#1044)
- \[pre-commit.ci\] pre-commit autoupdate (#1036)
- Run coverage upload in separate steep (#1053)
- Generate local report in `Tests` workflow and use proper script for fetch report (#1054)
- Move coverage back to main workflow (#1055)
- \[pre-commit.ci\] pre-commit autoupdate (#1056)
- \[pre-commit.ci\] pre-commit autoupdate (#1059)
- Update `actions/upload-artifact` and  `actions/download-artifact` from 3 to 4 (#1062)
- \[pre-commit.ci\] pre-commit autoupdate (#1064)
- Group actions update (#1065)
- \[pre-commit.ci\] pre-commit autoupdate (#1068)
- Remove requirement of 2 builds upload to codecov.io (#1073)
- Re add tests to coverage report (#1074)
- Switch from setup.cfg to pyproject.toml in workflows (#1076)
- Fix compiling pyinstaller pre-deps (#1075)
- Add codespell to pre-commit and fix pointed bugs (#1078)
- Add new ruff rules and apply them (#1079)
- \[pre-commit.ci\] pre-commit autoupdate (#1080)
- \[pre-commit.ci\] pre-commit autoupdate (#1081)
- Fix upgrade depenecies workflow (#1083)
- Block using `mpmath==1.4.0a0` and `sentry-sdk` 2.0.0a1/a2 in pre-test (#1085)
- \[pre-commit.ci\] pre-commit autoupdate (#1089)
- Fix jupyter failing test by using constraints (#1093)
- \[pre-commit.ci\] pre-commit autoupdate (#1091)
- \[pre-commit.ci\] pre-commit autoupdate (#1096)
- Add python 3.12 testing (#1087)
- Exclude pyside2 on python 3.11 and 3.12 from testing (#1099)
- \[pre-commit.ci\] pre-commit autoupdate (#1101)
- \[pre-commit.ci\] pre-commit autoupdate (#1103)
- Bump macos runners to macos-13 (both azure and GHA) (#1113)
- \[pre-commit.ci\] pre-commit autoupdate (#1108)
- Remove pyqt5 from constraints (#1118)

### 🛡️ Security

- *(deps)* Bump actions/checkout from 3 to 4 (#1029)
- *(deps)* Bump conda-incubator/setup-miniconda from 2 to 3 (#1038)
- *(deps)* Bump aganders3/headless-gui from 1 to 2 (#1047)
- *(deps)* Bump actions/checkout from 3 to 4 (#1045)
- *(deps)* Bump hynek/build-and-inspect-python-package from 1 to 2 (#1050)
- *(deps)* Bump actions/setup-python from 4 to 5 (#1046)
- *(deps)* Bump github/codeql-action from 2 to 3 (#1051)
- *(deps)* Bump peter-evans/create-pull-request from 5 to 6 (#1067)
- *(deps)* Bump codecov/codecov-action from 3 to 4 (#1066)

### Build

- Fix not bundling `Font Awesome 6 Free-Solid-900.otf` file to executable (#1114)
- Update readme and release to point to GitHub releases (#1115)

## 0.15.2 - 2023-08-28

### Bug Fixes

- Fix range threshold selection of algorithms (#1009)
- When run batch check if file extension is supported by loader (#1016)
- Do not allow to select and render corrupted batch plans (#1015)

### Testing

- \[Automatic\] Constraints upgrades: `imagecodecs`, `ipykernel`, `magicgui`, `psygnal`, `scipy`, `superqt`, `tifffile` (#1011)
- \[Automatic\] Constraints upgrades: `imageio`, `pyinstaller`, `tifffile` (#1018)

### Ci

- Use faster version of black (#1010)
- \[pre-commit.ci\] pre-commit autoupdate (#1013)

## 0.15.1 - 2023-08-09

### Bug Fixes

- Fix possible problem of double registration napari plugin in PartSeg bundle (#974)
- Bump OS versions for part of testing workflows.  (#977)
- Bump os version for main tests workflow. (#979)
- Ensure that the module `PartSegCore.channel_class` is present in bundle (#980)
- Lower npe2 schema version to work with older napari version (#981)
- Generate test report per platform (#978)
- Importing plugins in bundle keeping proper module names (#983)
- Fix napari repo workflow (#985)
- Fix bug in read tiff files with double `Q` in axes but one related to dummy dimension (#992)
- Fix bug that lead to corrupted state when saving calculation plan to excel file (#995)
- Enable python 3.11 test on CI, fix minor errors (#869)

### Features

- Allow to save multiple napari image layers to single tiff file (#1000)
- Add option to export batch project with data (#996)

### Testing

- \[Automatic\] Constraints upgrades: `imageio`, `ipython`, `psygnal`, `scipy`, `sentry-sdk` (#975)
- \[Automatic\] Constraints upgrades: `h5py`, `imagecodecs`, `imageio`, `ipykernel`, `napari`, `numpy`, `pandas`, `pydantic`, `pyinstaller`, `scipy`, `sentry-sdk`, `tifffile`, `vispy` (#986)
- \[Automatic\] Constraints upgrades: `imagecodecs`, `sentry-sdk`, `tifffile` (#997)
- \[Automatic\] Constraints upgrades: `ipykernel`, `pydantic` (#1002)
- \[Automatic\] Constraints upgrades: `numpy`, `pygments`, `sentry-sdk`, `superqt` (#1007)

### Ci

- \[pre-commit.ci\] pre-commit autoupdate (#973)
- \[pre-commit.ci\] pre-commit autoupdate (#982)
- \[pre-commit.ci\] pre-commit autoupdate (#987)
- \[pre-commit.ci\] pre-commit autoupdate (#988)
- \[pre-commit.ci\] pre-commit autoupdate (#991)
- \[pre-commit.ci\] pre-commit autoupdate (#998)
- \[pre-commit.ci\] pre-commit autoupdate (#1004)
- Change markdown linter from pre-commit to mdformat (#1006)

## 0.15.0 - 2023-05-30

### Bug Fixes

- Print all exceptions instead of the latest one in exception dialog (#799)
- Fix ROIExtractionResult `__str__`and `__repr__` to use `ROIExtractionResult` not `SegmentationResult` (#810)
- Fix code to address changes in the napari repository (#817)
- Fix the problem with resize of multiline widgets (#832)
- Fix tox configuration to run all required tests (#840)
- Fix MSO `step_limit` description in GUI (#843)
- Fix `redefined-while-unused`import code for python 3.9.7 (#844)
- Fix warnings reported by Deepsource (#846)
- Ensure that "ROI" layer is in the proper place for proper visualization (#856)
- Fix tests of napari widgets (#862)
- Fix build of the bundle for a new psygnal release (#863)
- Fix minimal requirements pipeline (#877)
- Fix bug with generation of form for the model with hidden field (#920)
- Update pyinstaller configuration (#926)
- Use text icon, not pixmap icon in colormap and labels list (#938)
- Resolve warnings when testing custom save dialog. (#941)
- Add padding zeros for component num when load Mask seg file to ROI GUI (#944)
- Proper calculate bounds for watershed napari widget (#969)
- Fix bug in the wrong order of axis saved in napari contribution (#972)

### Dependency

- \[Automatic\] Dependency upgrades (#824)
- \[Automatic\] Dependency upgrades (#828)
- \[Automatic\] Dependency upgrades: `ipykernel`, `packaging` (#838)
- \[Automatic\] Dependency upgrades: `imageio`, `ipykernel`, `napari`, `numpy`, `sentry` (#850)
- \[Automatic\] Dependency upgrades: `imagecodecs`, `ipykernel`, `numpy`, `psygnal` (#859)
- \[Automatic\] Dependency upgrades: `pydantic`, `pygments`, `xlsxwriter` (#874)
- \[Automatic\] Dependency upgrades: `imageio`, `packaging`, `scipy`, `xlsxwriter` (#878)
- \[Automatic\] Dependency upgrades: `ipykernel`, `requests`, `sentry`, `xlsxwriter` (#884)
- \[Automatic\] Dependency upgrades: `h5py`, `imagecodecs`, `imageio`, `ipykernel`, `pandas`, `sentry`, `tifffile` (#889)
- \[Automatic\] Dependency upgrades: `ipython`, `pyqt5` (#893)
- \[Automatic\] Dependency upgrades: `imageio`, `ipykernel`, `ipython`, `numpy`, `openpyxl`, `psygnal`, `pydantic`, `pyinstaller`, `pyqt5`, `scipy`, `sentry-sdk`, `tifffile`, `xlsxwriter` (#897)
- \[Automatic\] Dependency upgrades: `imageio`, `psygnal` (#905)
- \[Automatic\] Dependency upgrades: `ipython`, `magicgui`, `scipy`, `sentry-sdk`, `tifffile` (#906)
- \[Automatic\] Dependency upgrades: `imagecodecs`, `imageio`, `ipykernel`, `openpyxl`, `pydantic`, `pyinstaller`, `qtawesome`, `qtconsole`, `sentry-sdk`, `tifffile`, `xlsxwriter` (#908)
- \[Automatic\] Dependency upgrades: `imageio`, `ipykernel`, `ipython`, `pandas`, `psygnal`, `pydantic`, `pygments`, `pyinstaller`, `qtpy`, `sentry-sdk`, `tifffile` (#917)

### Documentation

- Update README and project metadata (#805)

### Features

- Add `PARTSEG_SENTRY_URL` env variable support and basic documentation about error reporting (#802)
- Allow to see underlying exception when show warning caused by exception (#829)
- Add voxel size measurement and allow to overwrite voxel size in batch (#853)
- Add alpha support for Qt6 (#866)
- Add option to create projection alongside z-axis (#919)
- Add napari image custom representation for better error report via sentry (#861)
- Add import and export operation for labels and colormaps (#936)
- Implement napari widgets for colormap and labels control (#935)
- Add forget all button to multiple files widget (#942)
- Do not abort processing whole mask segmentation project during exception on single component (#943)
- Add distance based watersheed to flow methods (#915)
- Add napari widgets for all group of algorithms (#958)
- Add napari widget to copy labels along z-axis (#968)

### Miscellaneous Tasks

- Improve ruff configuration, remove isort (#815)
- Use `fail_on_no_env` feature from `tox-gh-actions` (#842)
- Add python 3.11 to list of supported versions (#867)
- Disable python 3.11 test because of timeout (#870)
- Bump ruff to 0.0.218, remove flake8 from pre-commit (#880)
- Replace GabrielBB/xvfb-action@v1 by aganders3/headless-gui, part 2 (#887)
- Better minimal requirements test (#888)
- Improve regexp for proper generate list of packages in update report (#894)
- Add check for PR title (#933)
- Migrate from `nme` to `local_migrator` (#951)

### Refactor

- Simplify and refactor github workflows. (#864)
- Better load Mask project in Roi Analysis (#921)
- Use more descriptive names in `pylint: disable` (#922)
- Remove `pkg_resources` usage as it is deprecated (#967)
- Convert napari plugin to npe2 (#966)

### Security

- Bump peter-evans/create-pull-request from 4 to 5 (#928)

### Styling

- Change default theme to dark, remove blinking windows on startup. (#809)

### Testing

- \[Automatic\] Dependency upgrades: `packaging`, `pyinstaller`, `pyopengl-accelerate`, `tifffile`, `xlsxwriter` (#932)
- \[Automatic\] Constraints upgrades: `fonticon-fontawesome6`, `imageio`, `numpy`, `partsegcore-compiled-backend`, `pygments`, `sentry-sdk` (#937)
- \[Automatic\] Constraints upgrades: `imageio`, `ipython`, `pandas`, `requests`, `sentry-sdk` (#948)
- \[Automatic\] Constraints upgrades: `ipython`, `nme`, `qtconsole`, `requests`, `sentry-sdk` (#955)
- \[Automatic\] Constraints upgrades: `ipykernel`, `local-migrator`, `pyinstaller`, `sentry-sdk`, `sympy` (#957)
- \[Automatic\] Constraints upgrades: `sentry-sdk`, `xlsxwriter` (#959)
- \[Automatic\] Constraints upgrades: `requests` (#961)

### Ci

- Update codecov configuration to wait on two reports before post information (#934)
- \[pre-commit.ci\] pre-commit autoupdate (#945)
- \[pre-commit.ci\] pre-commit autoupdate (#956)
- \[pre-commit.ci\] pre-commit autoupdate (#964)

## 0.14.6 - 2022-11-13

### Bug Fixes

- Fix bug when loading already created project causing hide of ROI layer (#787)

### Features

- Improve error message if segmentation do not fit in ROI Mask (#788)

## 0.14.5 - 2022-11-09

### Bug Fixes

- Fix scalebar color (#774)
- Fix bug when saving segmentation parameters in mask analysis (#781)
- Fix multiple errors related to loading a new file in interactive mode (#784)

### Features

- Add an option for ensuring type in EventedDict and use it to validate profiles structures (#776)
- Add an option to create an issue from the error report dialog (#782)
- Add option for the multiline field in algorithm parameters (#766)

### Refactor

- Optimize CLI actions (#772)
- Clean warnings about threshold methods (#783)

### Build

- Bump chanzuckerberg/napari-hub-preview-action from 0.1.5 to 0.1.6 (#775)

## 0.14.4 - 2022-10-24

### Bug Fixes

- Fix `get_theme` calls to prepare for napari 0.4.17 (#729)
- Fix sentry tests (#742)
- Fix reporting error in load settings from the drive (#725)
- Fix saving pipeline from GUI (#756)
- Fix profile export/import dialogs (#761)
- Enable the "Compare" button if ROI is available (#765)
- Fix bug in cut with ROI to not make black artifacts (#767)

### Features

- Load alternatives labeling when opening PartSeg projects in napari (#731)
- Add option to toggle scale bar (#733)
- Allow customizing the settings directory using the `PARTSEG_SETTINGS_DIR` environment variable (#751)
- Separate recent algorithms from general application settings (#752)
- Add multiple otsu as threshold method with selection range of components (#710)
- Add function to load components from Mask Segmentation with a background in ROI Analysis (#768)

### Miscellaneous Tasks

- Prepare pyinstaller configuration for napari 0.4.17 (#748)
- Add ruff linter (#754)

### Testing

- Add new build and inspect wheel action (#747)

### Build

- Bump actions/checkout from 2 to 3 (#716)
- Bump actions/download-artifact from 1 to 3 (#709)

## 0.14.3 - 2022-08-18

### Bug Fixes

- Fix lack of rendering ROI when load image from segmentation (#694)
- Fix running ROI extraction from napari widget (#695)
- Delay setting image if an algorithm is still running (#627)
- Wrong error report when no component is found in restartable segmentation algorithm. (#633)
- Fix the process of building documentation (#653)

### Refactor

- Clean potential vulnerabilities  (#630)

### Testing

- Add more tests for common GUI elements  (#622)
- Report coverage per package. (#639)
- Update conda environment to not use PyQt5 in test (#646)
- Add tests files to calculate coverage (#655)

## 0.14.2 - 2022-05-05

### Bug Fixes

- Fix bug in save label colors between sessions (#610)
- Register PartSeg plugins before starting napari widgets. (#611)
- Mouse interaction with components works again after highlight. (#620)

### Refactor

- Limit test run (#603)
- Filter and solve warnings in tests (#607)
- Use QAbstractSpinBox.AdaptiveDecimalStepType in SpinBox instead of hardcoded bounds (#616)
- Clean and test `PartSeg.common_gui.universal_gui_part` (#617)

### Testing

- Speedup test by setup cache for pip (#604)
- Setup cache for azure pipelines workflows (#606)

## 0.14.1 - 2022-04-27

### Bug Fixes

- Update build wheels and sdist to have proper version tag (#583)
- Fix removing the first measurement entry in the napari Measurement widget (#584)
- Fix compatibility bug for conda Pyside2 version (#595)
- Error when synchronization is loaded, and newly loaded image has different dimensionality than currently loaded. (#598)

### Features

- Use pygments for coloring code in exception window (#591)
- Add option to calculate Measurement per Mask component (#590)

### Refactor

- Refactor the creation batch plan widgets and add tests for it (#587)
- Drop napari below 0.4.12 (#592)
- Update the order of ROI Mask algorithms to be the same as in older PartSeg versions (#600)

## 0.14.0 - 2022-04-14

### Bug Fixes

- Fix "Show selected" rendering mode in PartSeg ROI Mask (#565)
- Add access by operator `[]` to `pydantic.BaseModel` base structures for keeping backward compatibility (#579)

### Features

- Allow setting zoom factor from the interface in Search Label napari plugin (#538)
- Add controlling of zoom factor of search ROI in main GUI (#540)
- Better serialization mechanism allow for declaration data structure migration locally (#462)
- Make \`\*.obsep" file possible to load in PartSeg Analysis (#564)
- Add option to extract measurement profile or ROI extraction profile from the batch plan (#568)
- Allow import calculation plan from batch result excel file (#567)
- Improve error reporting when failing to deserialize data (#574)
- Launch PartSeg GUI from napari #581

### Refactor

- Store PartSegImage.Image channels as separated arrays (#554)
- Remove deprecated modules. (#429)
- Switch serialization backend to `nme` (#569)

### Testing

- Add test of creating AboutDialog (#539)
- Setup test for python 3.10. Disable `class_generator` test for this python (#570)

## 0.13.15

### Bug Fixes

- Using `translation` instead of `translation_grid` for shifting layers. (#474)
- Bugs in napari plugins (#478)
- Missing mask when using roi extraction from napari (#479)
- Fix segmentation fault on macos machines (#487)
- Fixes for napari 0.4.13 (#506)

### Documentation

- Create 0.13.15 release (#511)
- Add categories and preview page workflow for the napari hub (#489)

### Features

- Assign properties to mask layer in napari measurement widget (#480)

### Build

- Bump qtpy from 1.11.3 to 2.0.0 in /requirements (#498)
- Bump pydantic from 1.8.2 to 1.9.0 in /requirements (#496)
- Bump sentry-sdk from 1.5.1 to 1.5.2 in /requirements (#497)
- Bump sphinx from 4.3.1 to 4.3.2 in /requirements (#500)
- Bump pyinstaller from 4.7 to 4.8 in /requirements (#502)
- Bump pillow from 8.4.0 to 9.0.0 in /requirements (#501)
- Bump requests from 2.26.0 to 2.27.1 in /requirements (#495)
- Bump numpy from 1.21.4 to 1.22.0 in /requirements (#499)
- Bump numpy from 1.22.0 to 1.22.1 in /requirements (#509)
- Bump sphinx from 4.3.2 to 4.4.0 in /requirements (#510)

## 0.13.14

### Bug Fixes

- ROI alternative representation (#471)
- Change additive to translucent in rendering ROI and Mask (#472)

### Features

- Add morphological watershed segmentation (#469)
- Add Bilateral image filter (#470)

## 0.13.13

### Bug Fixes

- Fix bugs in the generation process of the changelog for release. (#428)
- Restoring ROI on home button click in compare viewer (#443)
- Fix Measurement name prefix in bundled PartSeg. (#458)
- Napari widgets registration in pyinstaller bundle (#465)
- Hide points button if no points are loaded, hide Mask checkbox if no mask is set (#463)
- Replace Label data instead of adding/removing layers - fix blending layers (#464)

### Features

- Add threshold information in layer annotation in the Multiple Otsu ROI extraction method (#430)
- Add option to select rendering method for ROI (#431)
- Add callback mechanism to ProfileDict, live update of ROI render parameters (#432)
- Move the info bar on the bottom of the viewer (#442)
- Add options to load recent files in multiple files widget (#444)
- Add ROI annotations as properties to napari labels layer created by ROI Extraction widgets (#445)
- Add signals to ProfileDict, remove redundant synchronization mechanisms (#449)
- Allow ignoring updates for 21 days (#453)
- Save all components if no components selected in mask segmentation (#456)
- Add modal dialog for search ROI components (#459)
- Add full measurement support as napari widget (#460)
- Add search labels as napari widget (#467)

### Refactor

- Export common code for load/save dialog to one place (#437)
- Change most of call QFileDialog to more generic code (#440)

### Testing

- Add test for `PartSeg.common_backend` module (#433)

## 0.13.12

### Bug Fixes

- Importing the previous version of settings (#406)
- Cutting without masking data (#407)
- Save in subdirectory in batch plan (#414)
- Loading plugins for batch processing (#423)

### Features

- Add randomization option for correlation calculation (#421)
- Add Imagej TIFF writer for image. (#405)
- Mask create widget for napari (#395)
- In napari roi extraction method show information from roi extraction method (#408)
- Add `*[0-9].tif` button in batch processing window (#412)
- Better label representation in 3d view (#418)

### Refactor

- Use Font Awesome instead of custom symbols (#424)

## 0.13.11

### Bug Fixes

- Adding mask in Prepare Plan for batch (#383)
- Set proper completion mode in SearchComboBox (#384)
- Showing warnings on the error with ROI load (#385)

### Features

- Add CellFromNucleusFlow "Cell from nucleus flow" cell segmentation method  (#367)
- When cutting components in PartSeg ROI mask allow not masking outer data (#379)
- Theme selection in GUI (#381)
- Allow return points from ROI extraction algorithm (#382)
- Add measurement to get ROI annotation by name. (#386)
- PartSeg ROI extraction algorithms as napari plugins (#387)
- Add  Pearson, Mander's, Intensity, Spearman colocalization measurements (#392)
- Separate standalone napari settings from PartSeg embedded napari settings (#397)

### Performance

- Use faster calc bound function (#375)

### Refactor

- Remove CustomApplication (#389)

## 0.13.10

- change tiff save backend to ome-tiff
- add `DistanceROIROI` and `ROINeighbourhoodROI` measurements

## 0.13.9

- annotation show bugfix

## 0.13.8

- napari deprecation fixes
- speedup simple measurement
- bundle plugins initial support

## 0.13.7

- add measurements widget for napari
- fix bug in pipeline usage

## 0.13.6

- Hotfix release
- Prepare for a new napari version

## 0.13.5

- Small fixes for error reporting
- Fix mask segmentation

## 0.13.4

- Bugfix for outdated profile/pipeline preview

## 0.13.3

- Fix saving roi_info in multiple files and history

## 0.13.2

- Fix showing label in select label tab

## 0.13.1

- Add Haralick measurements
- Add obsep file support

## 0.13.0

- Add possibility of custom input widgets for algorithms
- Switch to napari Colormaps instead of custom one
- Add points visualization
- Synchronization widget for builtin (View menu) napari viewer
- Drop Python 3.6

## 0.12.7

- Fixes for napari 0.4.6

## 0.12.6

- Fix prev_mask_get
- Fix cache mechanism on mask change
- Update PyInstaller build

## 0.12.5

- Fix bug in pipeline execute

## 0.12.4

- Fix ROI Mask windows related build (signal not properly connected)

## 0.12.3

- Fix ROI Mask

## 0.12.2

- Fix windows bundle

## 0.12.1

- History of last opened files
- Add ROI annotation and ROI alternatives
- Minor bugfix

## 0.12.0

- Toggle multiple files widget in View menu
- Toggle Left panel in ROI Analysis in View Menu
- Rename Mask Segmentation to ROI Mask
- Add documentation for interface
- Add Batch processing tutorial
- Add information about errors to batch processing output file
- Load image from the batch prepare window
- Add search option in part of list and combo boxes
- Add drag and drop mechanism to load list of files to batch window.

## 0.11.5

- add side view to viewer
- fix horizontal view for Measurements result table

## 0.11.4

- bump to napari 0.3.8 in bundle
- fix bug with not presented segmentation loaded from project
- add frame (1 pix) to image cat from base one based on segmentation
- pin to Qt version to 5.14

## 0.11.3

- prepare for napari 0.3.7
- split napari io plugin on multiple part
- better reporting for numpy array via sentry
- fix setting color for mask marking

## 0.11.2

- Speedup image set in viewer using async calls
- Fix bug in long name of sheet with parameters

## 0.11.1

- Add screenshot option in View menu
- Add Voxels measurements

## 0.11.0

- Make sprawl algorithm name shorter
- Unify capitalisation of measurement names
- Add simple measurements to mask segmentation
- Use napari as viewer
- Add possibility to preview additional output of algorithms (In View menu)
- Update names of available Algorithm and Measurement to be more descriptive.

## 0.10.8

- fix synchronisation between viewers in Segmentation Analysis
- fix batch crash on error during batch run, add information about file on which calculation fails
- add changelog preview in Help > About

## 0.10.7

- in measurements, on empty list of components mean will return 0

## 0.10.6

- fix border rim preview
- fix problem with size of image preview
- zoom with scroll and moving if rectangle zoom is not marked

## 0.10.5

- make PartSeg PEP517 compatible.
- fix multiple files widget on Windows (path normalisation)

## 0.10.4

- fix slow zoom

## 0.10.3

- deterministic order of elements in batch processing.

## 0.10.2

- bugfixes

## 0.10.1

- bugfixes

## 0.10.0

- Add creating custom label coloring.
- Change execs interpreter to python 3.7.
- Add masking operation in Segmentation Mask.
- Change license to BSD.
- Allow select root type in batch processing.
- Add median filter in preview.

## 0.9.7

- fix bug in compare mask

## 0.9.6

- fix bug in loading project with mask
- upgrade PyInstaller version (bug  GHSA-7fcj-pq9j-wh2r)

## 0.9.5

- fix bug in loading project in "Segmentation analysis"

## 0.9.4

- read mask segmentation projects
- choose source type in batch
- add initial support to OIF and CZI file format
- extract utils to PartSegCore module
- add automated tests of example notebook
- reversed mask
- load segmentation parameters in mask segmentation
- allow use sprawl in segmentation tool
- add radial split of mask for measurement
- add all measurement results in batch, per component sheet

## 0.9.3

- start automated build documentation
- change color map backend and allow for user to create custom color map.
- segmentation compare
- update test engines
- support of PySide2

## 0.9.2.3

- refactor code to make easier create plugin for mask segmentation
- create class base updater for update outdated algorithm description
- fix save functions
- fix different bugs

## 0.9.2.2

- extract static data to separated package
- update marker of fix range and add mark of gauss in channel control

## 0.9.2.1

- add VoteSmooth and add choosing of smooth algorithm

## 0.9.2

- add pypi base check for update

- remove resetting image state when change state in same image

- in stack segmentation add options to picking components from segmentation's

- in mask segmentation add:

  - preview of segmentation parameters per component,
  - save segmentation parameters in save file
  - new implementation of batch mode.

## 0.9.1

- Add multiple files widget

- Add Calculating distances between segmented object and mask

- Batch processing plan fixes:

  - Fix adding pipelines to plan
  - Redesign mask widget

- modify measurement backend to allow calculate multi channel measurements.

## 0.9

Begin of changelog
"""



from PartSegCore.algorithm_describe_base import Register
from PartSegCore.image_transforming.image_projection import ImageProjection
from PartSegCore.image_transforming.interpolate_image import InterpolateImage
from PartSegCore.image_transforming.swap_time_stack import SwapTimeStack
from PartSegCore.image_transforming.transform_base import TransformBase

image_transform_dict = Register(InterpolateImage, SwapTimeStack, ImageProjection)

__all__ = ("image_transform_dict", "InterpolateImage", "TransformBase", "ImageProjection")



import multiprocessing
from importlib.metadata import version as package_version

import numpy as np
import pytest
import sentry_sdk
import sentry_sdk.serializer
import sentry_sdk.utils
from packaging.version import parse as parse_version
from sentry_sdk.client import Client
from sentry_sdk.hub import Hub
from sentry_sdk.serializer import serialize

from PartSegCore.analysis.batch_processing.batch_backend import prepare_error_data
from PartSegCore.utils import safe_repr

SENTRY_GE_1_29 = parse_version(package_version("sentry_sdk")) >= parse_version("1.29.0")

if SENTRY_GE_1_29:
    DEFAULT_ERROR_REPORT = sentry_sdk.utils.DEFAULT_MAX_VALUE_LENGTH
    CONST_NAME = "DEFAULT_MAX_VALUE_LENGTH"
else:
    DEFAULT_ERROR_REPORT = sentry_sdk.utils.MAX_STRING_LENGTH
    CONST_NAME = "MAX_STRING_LENGTH"


def test_message_clip(monkeypatch):
    message = "a" * 5000
    assert len(sentry_sdk.utils.strip_string(message).value) == DEFAULT_ERROR_REPORT
    monkeypatch.setattr(sentry_sdk.utils, CONST_NAME, 10**4)
    assert len(sentry_sdk.utils.strip_string(message)) == 5000


def test_sentry_serialize_clip(monkeypatch):
    message = "a" * 5000
    try:
        raise ValueError("eeee")
    except ValueError as e:
        event, _hint = sentry_sdk.utils.event_from_exception(e)
        event["message"] = message

        clipped = serialize(event)
        assert len(clipped["message"]) == DEFAULT_ERROR_REPORT
        monkeypatch.setattr(sentry_sdk.utils, CONST_NAME, 10**4)
        clipped = serialize(event)
        assert len(clipped["message"]) == 5000


def test_sentry_variables_clip(monkeypatch):
    letters = "abcdefghijklmnoprst"
    for letter in letters:
        locals()[letter] = 1
    try:
        raise ValueError("eeee")
    except ValueError as ee:
        event, hint = sentry_sdk.utils.event_from_exception(ee)
        vars_dict = event["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]
        for letter in letters:
            assert letter in vars_dict
        clipped = serialize(event)
        assert len(clipped["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]) == 10
        monkeypatch.setattr(sentry_sdk.serializer, "MAX_DATABAG_BREADTH", 100)
        clipped = serialize(event)
        assert len(clipped["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]) == len(vars_dict)
        assert len(clipped["exception"]["values"][0]["stacktrace"]["frames"][0]["vars"]) > 10

        client = Client("https://aaa@test.pl/77")
        Hub.current.bind_client(client)
        sentry_sdk.capture_event(event, hint=hint)


def test_sentry_report(monkeypatch):
    message = "a" * 5000
    happen = [False]

    def check_event(event):
        happen[0] = True
        assert len(event["message"]) == DEFAULT_ERROR_REPORT
        assert len(event["extra"]["lorem"]) == DEFAULT_ERROR_REPORT

    def check_envelope(envelope):
        check_event(envelope.get_event())

    try:
        raise ValueError("eeee")
    except ValueError as e:
        event, hint = sentry_sdk.utils.event_from_exception(e)
        event["message"] = message
        client = Client("https://aaa@test.pl/77")
        Hub.current.bind_client(client)
        monkeypatch.setattr(client.transport, "capture_event", check_event)
        # handle sentry-sdk 2
        monkeypatch.setattr(client.transport, "capture_envelope", check_envelope, raising=False)

        with sentry_sdk.push_scope() as scope:
            scope.set_extra("lorem", message)
            sentry_sdk.capture_event(event, hint=hint)
        assert happen[0] is True


def test_sentry_report_no_clip(monkeypatch):
    message = "a" * 5000
    happen = [False]
    monkeypatch.setattr(sentry_sdk.utils, CONST_NAME, 10**4)

    def check_event(event):
        happen[0] = True
        assert len(event["message"]) == 5000
        assert len(event["extra"]["lorem"]) == 5000

    def check_envelope(envelope):
        check_event(envelope.get_event())

    try:
        raise ValueError("eeee")
    except ValueError as e:
        event, hint = sentry_sdk.utils.event_from_exception(e)
        event["message"] = message
        if SENTRY_GE_1_29:
            client = Client("https://aaa@test.pl/77", max_value_length=10**4)
        else:
            client = Client("https://aaa@test.pl/77")
        Hub.current.bind_client(client)
        monkeypatch.setattr(client.transport, "capture_event", check_event)
        # handle sentry-sdk 2
        monkeypatch.setattr(client.transport, "capture_envelope", check_envelope, raising=False)
        with sentry_sdk.push_scope() as scope:
            scope.set_extra("lorem", message)
            event_id = sentry_sdk.capture_event(event, hint=hint)
        assert event_id is not None
        assert happen[0] is True


def exception_fun(num: int):
    if num < 1:
        raise ValueError("test")
    exception_fun(num - 1)


def executor_fun(que: multiprocessing.Queue):
    try:
        exception_fun(10)
    except ValueError as e:
        ex, (event, tr) = prepare_error_data(e)
        que.put((ex, event, tr))


def test_exception_pass(monkeypatch):
    def check_event(event):
        assert len(event["exception"]["values"][0]["stacktrace"]["frames"]) == 12

    message_queue = multiprocessing.get_context("spawn").Queue()
    p = multiprocessing.get_context("spawn").Process(target=executor_fun, args=(message_queue,))
    p.start()
    p.join()
    assert not message_queue.empty()
    ex, event, _tr = message_queue.get()
    assert isinstance(ex, ValueError)
    assert isinstance(event, dict)
    client = Client("https://aaa@test.pl/77")
    Hub.current.bind_client(client)
    monkeypatch.setattr(client.transport, "capture_event", check_event)
    event_id = sentry_sdk.capture_event(event)
    assert event_id is not None


@pytest.mark.parametrize("dtype", [np.uint8, np.int8, np.float32])
def test_numpy_array_serialize(monkeypatch, dtype):
    arr = np.zeros((10, 10), dtype=dtype)
    arr[1, 5] = 10
    monkeypatch.setattr(sentry_sdk.serializer, "safe_repr", safe_repr)
    res = serialize(arr)
    assert res == f"array(size={arr.size}, shape={arr.shape}, dtype={arr.dtype}, min={arr.min()}, max={arr.max()})"


def test_small_numpy_serialize(monkeypatch):
    arr = np.zeros(10, dtype=np.uint8)
    arr[1] = 10
    monkeypatch.setattr(sentry_sdk.serializer, "safe_repr", safe_repr)
    res = serialize(arr)
    assert res == str(res)



import typing
from io import BytesIO
from pathlib import Path

import SimpleITK

from PartSegCore.analysis.save_functions import ProjectTuple
from PartSegCore.io_utils import SaveBase


class SaveITKSnap(SaveBase):
    @classmethod
    def get_name(cls):
        return "Mask for itk-snap (*.img)"

    @classmethod
    def get_short_name(cls):
        return "img"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info: ProjectTuple,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        mask = SimpleITK.GetImageFromArray(project_info.roi_info.roi)
        SimpleITK.WriteImage(mask, save_location)


def register():
    from PartSegCore.register import RegisterEnum
    from PartSegCore.register import register as register_fun

    register_fun(SaveITKSnap, RegisterEnum.analysis_save)



#!/usr/bin/env python


#############################################################################
#
# Copyright (C) 2013 Riverbank Computing Limited.
# Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
# All rights reserved.
#
# This file is part of the examples of PyQt.
#
# $QT_BEGIN_LICENSE:BSD$
# You may use this file under the terms of the BSD license as follows:
#
# "Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#   * Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#   * Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in
#     the documentation and/or other materials provided with the
#     distribution.
#   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
#     the names of its contributors may be used to endorse or promote
#     products derived from this software without specific prior written
#     permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
# $QT_END_LICENSE$
#
#############################################################################

from qtpy.QtCore import QPoint, QRect, QSize, Qt
from qtpy.QtWidgets import QLayout, QSizePolicy


class FlowLayout(QLayout):
    def __init__(self, parent=None, margin=0, spacing=-1):
        super().__init__(parent)

        if parent is not None:
            self.setContentsMargins(margin, margin, margin, margin)

        self.setSpacing(spacing)

        self.itemList = []

    def __del__(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def clear(self):
        item = self.takeAt(0)
        while item:
            item = self.takeAt(0)

    def addItem(self, item):
        self.itemList.append(item)

    def addLayout(self, layout):
        self.itemList.append(layout)

    def addStretch(self):
        """To satisfy the QLayout API"""

    def count(self):
        return len(self.itemList)

    def itemAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList[index]

        return None

    def takeAt(self, index):
        if 0 <= index < len(self.itemList):
            return self.itemList.pop(index)
        return None

    def insertWidget(self, index, widget):
        self.addWidget(widget)
        i = self.indexOf(widget)
        element = self.itemList.pop(i)
        self.itemList.insert(index, element)

    def expandingDirections(self):  # pylint: disable=no-self-use
        return Qt.Orientation(0)

    def hasHeightForWidth(self):  # pylint: disable=no-self-use
        return True

    def heightForWidth(self, width):
        return self.doLayout(QRect(0, 0, width, 0), True)

    def setGeometry(self, rect):
        super().setGeometry(rect)
        self.doLayout(rect, False)

    def sizeHint(self):
        return self.minimumSize()

    def minimumSize(self):
        size = QSize()

        for item in self.itemList:
            size = size.expandedTo(item.minimumSize())

        margin, _, _, _ = self.getContentsMargins()

        size += QSize(2 * margin, 2 * margin)
        return size

    def doLayout(self, rect, testOnly):
        x = rect.x()
        y = rect.y()
        line_height = 0

        for item in self.itemList:
            wid = item.widget()
            if wid is not None:
                space_x = self.spacing() + wid.style().layoutSpacing(
                    QSizePolicy.ControlType.PushButton, QSizePolicy.ControlType.PushButton, Qt.Orientation.Horizontal
                )
                space_y = self.spacing() + wid.style().layoutSpacing(
                    QSizePolicy.ControlType.PushButton, QSizePolicy.ControlType.PushButton, Qt.Orientation.Vertical
                )
            else:
                space_x = self.spacing()  # + wid.layoutSpacing(QSizePolicy.PushButton, QSizePolicy.PushButton,
                space_y = self.spacing()  # + wid.layoutSpacing(QSizePolicy.PushButton, QSizePolicy.PushButton,
            next_x = x + item.sizeHint().width() + space_x
            if next_x - space_x > rect.right() and line_height > 0:
                x = rect.x()
                y = y + line_height + space_y
                next_x = x + item.sizeHint().width() + space_x
                line_height = 0

            if not testOnly:
                item.setGeometry(QRect(QPoint(x, y), item.sizeHint()))

            x = next_x
            line_height = max(line_height, item.sizeHint().height())

        return y + line_height - rect.y()



import os.path
import typing
from abc import abstractmethod
from contextlib import suppress
from io import BytesIO
from pathlib import Path
from threading import Lock

import numpy as np
import tifffile
from czifile.czifile import CziFile
from defusedxml import ElementTree
from oiffile import OifFile

from PartSegImage.image import Image

INCOMPATIBLE_IMAGE_MASK = "Incompatible shape of mask and image"


if typing.TYPE_CHECKING:
    from xml.etree.ElementTree import Element  # nosec


def _empty(_, __):
    """Empty function for callback"""


class TiffFileException(Exception):
    """
    exception raised if reading tiff file fails. Created for distinguish exceptions which should
    reported as warning message (not for report)
    """


class BaseImageReader:
    """
    Base class for reading image using Christopher Gholike libraries

    :cvar typing.Type[Image] ~.image_class: image class to return
    """

    image_class = Image

    @classmethod
    def return_order(cls) -> str:
        """
        Order to which image axes should be rearranged before pass to :py:attr:`image_class` constructor.
        Default is :py:attr:`image_class.return_order`
        """
        return cls.image_class.axis_order

    def __init__(self, callback_function=None):
        self.default_spacing = 10**-6, 10**-6, 10**-6
        self.spacing = self.default_spacing
        if callback_function is None:
            self.callback_function = _empty
        else:
            self.callback_function = callback_function

    def set_default_spacing(self, spacing):
        spacing = tuple(spacing)
        if len(spacing) == 2:
            # one micrometer
            spacing = (10 ** (-6), *spacing)
        if len(spacing) != 3:
            raise ValueError(f"wrong spacing {spacing}")  # pragma: no cover
        self.default_spacing = spacing

    @abstractmethod
    def read(self, image_path: typing.Union[str, Path], mask_path=None, ext=None) -> Image:
        """
        Main function to read image. If ext is not set then it may be deduced from path to file.
        If BytesIO is given and non default data file type is needed then ext need to be set

        :param image_path: path to image or buffer
        :param mask_path: path to mask or buffer
        :param ext: extension to decide algorithm, if absent and image_path is path then
            should be deduced from path
        :return: image structure
        """
        raise NotImplementedError

    @classmethod
    def read_image(
        cls,
        image_path: typing.Union[str, Path],
        mask_path=None,
        callback_function: typing.Optional[typing.Callable] = None,
        default_spacing: typing.Optional[typing.Tuple[float, float, float]] = None,
    ) -> Image:
        """
        read image file with optional mask file

        :param image_path: path or opened file contains image
        :param mask_path:
        :param callback_function: function for provide information about progress in reading file (for progressbar)
        :param default_spacing: used if file do not contains information about spacing
            (or metadata format is not supported)
        :return: image
        """
        # TODO add generic description of callback function
        instance = cls(callback_function)
        if default_spacing is not None:
            instance.set_default_spacing(default_spacing)
        return instance.read(image_path, mask_path)

    @staticmethod
    def _reduce_obsolete_dummy_axes(array, axes) -> typing.Tuple[np.ndarray, str]:
        """
        If there are duplicates in axes string then remove dimensions of size one

        :return: reduced array and axes
        """
        if len(axes) == len(set(axes)):
            return array, axes

        ax_li = []
        shape_li = []
        for dim, ax in zip(array.shape, axes):
            if dim != 1:
                ax_li.append(ax)
                shape_li.append(dim)
        axes = "".join(ax_li)
        array = np.reshape(array, shape_li)
        return array, axes

    @classmethod
    def update_array_shape(cls, array: np.ndarray, axes: str):
        """
        Rearrange order of array axes to get proper internal axes order

        :param array: array to reorder
        :param axes: current order of array axes as string like "TZYXC"
        """
        array, axes = cls._reduce_obsolete_dummy_axes(array, axes)

        try:
            final_mapping_dict = {letter: i for i, letter in enumerate(cls.return_order())}
            for let1, let2 in [("Z", "I"), ("Z", "Q"), ("C", "S")]:
                if let1 in final_mapping_dict and let2 not in final_mapping_dict:
                    final_mapping_dict[let2] = final_mapping_dict[let1]
            axes_li = list(axes)
            # Fixme; workaround for old saved segmentation
            if axes_li[0] == "Q" and axes_li[1] == "Q":
                axes_li[0] = "T"
                axes_li[1] = "Z"
            i = 0
            while i < len(axes_li):
                name = axes_li[i]
                if name not in final_mapping_dict and array.shape[i] == 1:
                    array = array.take(0, i)
                    axes_li.pop(i)
                else:
                    i += 1

            final_mapping = [final_mapping_dict[letter] for letter in axes_li]
        except KeyError as e:  # pragma: no cover
            raise NotImplementedError(
                f"Data type not supported ({e.args[0]}). Please contact with author for update code"
            ) from e
        if len(final_mapping) != len(set(final_mapping)):  # pragma: no cover
            raise NotImplementedError("Data type not supported. Please contact with author for update code")
        if len(array.shape) < len(cls.return_order()):
            array = np.reshape(array, array.shape + (1,) * (len(cls.return_order()) - len(array.shape)))

        return np.moveaxis(array, list(range(len(axes_li))), final_mapping)


class BaseImageReaderBuffer(BaseImageReader):
    @abstractmethod
    def read(self, image_path: typing.Union[str, Path, BytesIO], mask_path=None, ext=None) -> Image:
        """
        Main function to read image. If ext is not set then it may be deduced from path to file.
        If BytesIO is given and non default data file type is needed then ext need to be set

        :param image_path: path to image or buffer
        :param mask_path: path to mask or buffer
        :param ext: extension if need to decide algorithm, if absent and image_path is path then
            should be deduced from path
        :return: image structure
        """
        raise NotImplementedError

    @classmethod
    def read_image(
        cls,
        image_path: typing.Union[str, Path, BytesIO],
        mask_path=None,
        callback_function: typing.Optional[typing.Callable] = None,
        default_spacing: typing.Optional[typing.Tuple[float, float, float]] = None,
    ) -> Image:
        """
        read image file with optional mask file

        :param image_path: path or opened file contains image
        :param mask_path:
        :param callback_function: function for provide information about progress in reading file (for progressbar)
        :param default_spacing: used if file do not contains information about spacing
            (or metadata format is not supported)
        :return: image
        """
        # TODO add generic description of callback function
        instance = cls(callback_function)
        if default_spacing is not None:
            instance.set_default_spacing(default_spacing)
        return instance.read(image_path, mask_path)


class GenericImageReader(BaseImageReaderBuffer):
    """This class try to decide which method use base on path"""

    def read(self, image_path: typing.Union[str, BytesIO, Path], mask_path=None, ext=None) -> Image:
        if ext is None:
            if isinstance(image_path, (str, Path)):
                ext = os.path.splitext(image_path)[1]
            else:
                ext = ".tif"
        ext = ext.lower()
        if ext == ".czi":
            return CziImageReader.read_image(image_path, mask_path, self.callback_function, self.default_spacing)
        if ext in [".oif", ".oib"]:
            if isinstance(image_path, BytesIO):
                raise NotImplementedError("Oif format is not supported for BytesIO")
            return OifImagReader.read_image(image_path, mask_path, self.callback_function, self.default_spacing)
        if ext == ".obsep":
            if isinstance(image_path, BytesIO):
                raise NotImplementedError("Obsep format is not supported for BytesIO")
            return ObsepImageReader.read_image(image_path, mask_path, self.callback_function, self.default_spacing)
        return TiffImageReader.read_image(image_path, mask_path, self.callback_function, self.default_spacing)


class OifImagReader(BaseImageReader):
    def read(self, image_path: typing.Union[str, Path], mask_path=None, ext=None) -> Image:
        with OifFile(image_path) as image_file:
            tiffs = tifffile.natural_sorted(image_file.glob("*.tif"))
            with tifffile.TiffFile(image_file.open_file(tiffs[0]), name=tiffs[0]) as tif_file:
                axes = image_file.series[0].axes + tif_file.series[0].axes
            image_data = image_file.asarray()
            image_data = self.update_array_shape(image_data, axes)
            with suppress(KeyError):
                self._read_scale_parameter(image_file)
                # TODO add mask reading
        return self.image_class(
            image_data, self.spacing, file_path=os.path.abspath(image_path), axes_order=self.return_order()
        )

    def _read_scale_parameter(self, image_file):
        flat_parm = image_file.mainfile["Reference Image Parameter"]
        x_scale = flat_parm["HeightConvertValue"] * name_to_scalar[flat_parm["HeightUnit"]]
        y_scale = flat_parm["WidthConvertValue"] * name_to_scalar[flat_parm["WidthUnit"]]
        i = 0
        while True:
            name = f"Axis {i} Parameters Common"
            if name not in image_file.mainfile:  # pragma: no cover
                z_scale = 1
                break
            axis_info = image_file.mainfile[name]
            if axis_info["AxisCode"] == "Z":
                z_scale = axis_info["Interval"] * name_to_scalar[axis_info["UnitName"]]
                break
            i += 1

        self.spacing = z_scale, x_scale, y_scale


class CziImageReader(BaseImageReaderBuffer):
    """
    This class is to read data from czi files. Masks will be treated as TIFF.
    """

    def read(self, image_path: typing.Union[str, BytesIO, Path], mask_path=None, ext=None) -> Image:
        image_file = CziFile(image_path)
        image_data = image_file.asarray()
        image_data = self.update_array_shape(image_data, image_file.axes)
        metadata = image_file.metadata(False)
        with suppress(KeyError):
            scaling = metadata["ImageDocument"]["Metadata"]["Scaling"]["Items"]["Distance"]
            scale_info = {el["Id"]: el["Value"] for el in scaling}
            self.spacing = (
                scale_info.get("Z", self.default_spacing[0]),
                scale_info.get("Y", self.default_spacing[1]),
                scale_info.get("X", self.default_spacing[2]),
            )
        # TODO add mask reading
        if isinstance(image_path, BytesIO):
            image_path = ""
        return self.image_class(image_data, self.spacing, file_path=image_path, axes_order=self.return_order())

    @classmethod
    def update_array_shape(cls, array: np.ndarray, axes: str):
        if "B" in axes:
            index = axes.index("B")
            if array.shape[index] != 1:
                raise NotImplementedError(
                    "Czi file with B axes is not currently supported by PartSeg."
                    " Please contact with author for update code"
                )
            array = array.take(0, axis=index)
            axes = axes[:index] + axes[index + 1 :]
        if axes[-1] == "0":
            array = array[..., 0]
            axes = axes[:-1]
        return super().update_array_shape(array, axes)


class ObsepImageReader(BaseImageReader):
    def _search_for_files(
        self,
        directory: Path,
        channels: typing.List["Element"],
        suffix: str = "",
        required: bool = False,
    ) -> typing.List[Image]:
        possible_extensions = [".tiff", ".tif", ".TIFF", ".TIF"]
        channel_list = []
        for channel in channels:
            try:
                name = next(iter(channel)).attrib["val"] + suffix
            except StopIteration as e:  # pragma: no cover
                raise ValueError("Missed information about channel name in obsep file") from e
            for ex in possible_extensions:
                if (directory / (name + ex)).exists():
                    name += ex
                    break
            else:  # pragma: no cover
                if required:
                    raise ValueError(f"Not found file for key {name}")
                continue
            channel_list.append(TiffImageReader.read_image(directory / name, default_spacing=self.default_spacing))
        return channel_list

    def read(self, image_path: typing.Union[str, Path], mask_path=None, ext=None) -> Image:
        directory = Path(os.path.dirname(image_path))
        xml_doc = ElementTree.parse(image_path).getroot()
        channels = xml_doc.findall("net/node/node/attribute[@name='image type']")
        if not channels:
            raise ValueError("Information about channel images not found")  # pragma: no cover
        channel_list = [
            *self._search_for_files(directory, channels, required=True),
            *self._search_for_files(directory, channels, "_deconv"),
        ]
        image = channel_list[0]
        for el in channel_list[1:]:
            image = image.merge(el, "C")

        z_spacing = (
            float(xml_doc.find("net/node/attribute[@name='step width']/double").attrib["val"]) * name_to_scalar["um"]
        )

        image.set_spacing((z_spacing,) + image.spacing[1:])
        image.file_path = str(image_path)
        return image


class TiffImageReader(BaseImageReaderBuffer):
    """
    TIFF/LSM files reader. Base reading with :py:meth:`BaseImageReader.read_image`

    image_file: tifffile.TiffFile
    mask_file: tifffile.TiffFile
    """

    def __init__(self, callback_function=None):
        super().__init__(callback_function)
        self.colors = None
        self.channel_names = None
        self.ranges = None
        self.shift = (0, 0, 0)
        self.name = ""

    def read(self, image_path: typing.Union[str, BytesIO, Path], mask_path=None, ext=None) -> Image:
        """
        Read tiff image from tiff_file
        """
        self.spacing, self.colors, self.channel_names, self.ranges = self.default_spacing, None, None, None
        with tifffile.TiffFile(image_path) as image_file:
            total_pages_num = len(image_file.series[0])

            axes = image_file.series[0].axes

            if image_file.is_lsm:
                self.read_lsm_metadata(image_file)
            elif image_file.is_imagej:
                self.read_imagej_metadata(image_file)
            elif image_file.is_ome:
                self.read_ome_metadata(image_file)
            else:
                x_spacing, y_spacing = self.read_resolution_from_tags(image_file)
                self.spacing = self.default_spacing[0], y_spacing, x_spacing
            mutex = Lock()
            count_pages = [0]

            def report_func():
                mutex.acquire()
                count_pages[0] += 1
                self.callback_function("step", count_pages[0])
                mutex.release()

            if mask_path is not None:
                with tifffile.TiffFile(mask_path) as mask_file:
                    self.callback_function("max", total_pages_num + len(mask_file.series[0]))
                    self.verify_mask(mask_file, image_file)
                    mask_file.report_func = report_func
                    mask_data = mask_file.asarray()
                    mask_data = self.update_array_shape(mask_data, mask_file.series[0].axes)
                    if "C" in self.return_order():
                        pos: typing.List[typing.Union[slice, int]] = [slice(None) for _ in range(mask_data.ndim)]
                        pos[self.return_order().index("C")] = 0
                        mask_data = mask_data[tuple(pos)]

            else:
                mask_data = None
                if total_pages_num > 1:
                    self.callback_function("max", total_pages_num)

            image_file.report_func = report_func
            try:
                image_data = image_file.asarray()
            except ValueError as e:  # pragma: no cover
                raise TiffFileException(*e.args) from e
            image_data = self.update_array_shape(image_data, axes)

        if not isinstance(image_path, (str, Path)):
            image_path = ""
        return self.image_class(
            image_data,
            self.spacing,
            mask=mask_data,
            default_coloring=self.colors,
            channel_names=self.channel_names,
            ranges=self.ranges,
            file_path=os.path.abspath(image_path),
            axes_order=self.return_order(),
            shift=self.shift,
            name=self.name,
        )

    @staticmethod
    def verify_mask(mask_file, image_file):
        """
        verify if mask fit to image. Raise ValueError exception on error
        :return:
        """
        if mask_file is None:  # pragma: no cover
            return
        image_series = image_file.pages[0]
        mask_series = mask_file.pages[0]
        for i, pos in enumerate(mask_series.axes):
            if mask_series.shape[i] == 1:  # pragma: no cover
                continue
            try:
                j = image_series.axes.index(pos)
            except ValueError as e:  # pragma: no cover
                raise ValueError(f"{INCOMPATIBLE_IMAGE_MASK} (axes)") from e
                # TODO add verification if problem with T/Z/I
            if image_series.shape[j] != mask_series.shape[i]:  # pragma: no cover
                raise ValueError(INCOMPATIBLE_IMAGE_MASK)
            # TODO Add verification if mask have to few dimensions

    @staticmethod
    def decode_int(val: int):
        """
        This function split 32 bits int on 4 8-bits ints

        :param val: value to decode
        :return: list of four numbers with values from [0, 255]
        """
        return [(val >> x) & 255 for x in [24, 16, 8, 0]]

    def read_resolution_from_tags(self, image_file):
        tags = image_file.pages[0].tags
        try:
            if image_file.is_imagej:
                scalar = name_to_scalar[image_file.imagej_metadata["unit"]]
            else:
                unit = tags["ResolutionUnit"].value
                if unit == 3:
                    scalar = name_to_scalar["centimeter"]
                elif unit == 2:
                    scalar = name_to_scalar["cal"]
                else:  # pragma: no cover
                    raise KeyError(f"wrong scalar {tags['ResolutionUnit']}, {tags['ResolutionUnit'].value}")

            x_spacing = tags["XResolution"].value[1] / tags["XResolution"].value[0] * scalar
            y_spacing = tags["YResolution"].value[1] / tags["YResolution"].value[0] * scalar
        except (KeyError, ZeroDivisionError):
            x_spacing, y_spacing = self.default_spacing[2], self.default_spacing[1]
        return x_spacing, y_spacing

    def read_imagej_metadata(self, image_file):
        try:
            z_spacing = image_file.imagej_metadata["spacing"] * name_to_scalar[image_file.imagej_metadata["unit"]]
        except KeyError:
            z_spacing = self.default_spacing[0]
        x_spacing, y_spacing = self.read_resolution_from_tags(image_file)
        self.spacing = z_spacing, y_spacing, x_spacing
        self.colors = image_file.imagej_metadata.get("LUTs")
        self.channel_names = image_file.imagej_metadata.get("Labels")
        if "Ranges" in image_file.imagej_metadata:
            ranges = image_file.imagej_metadata["Ranges"]
            self.ranges = list(zip(ranges[::2], ranges[1::2]))

    def _read_ome_channel_information(self, meta_data):
        if "Channel" not in meta_data["Pixels"]:
            return
        if isinstance(meta_data["Pixels"]["Channel"], (list, tuple)):
            with suppress(KeyError):
                self.channel_names = [ch["Name"] for ch in meta_data["Pixels"]["Channel"]]
            with suppress(KeyError):
                self.colors = [self.decode_int(ch["Color"])[:-1] for ch in meta_data["Pixels"]["Channel"]]
            return
        if "Name" in meta_data["Pixels"]["Channel"]:
            self.channel_names = [meta_data["Pixels"]["Channel"]["Name"]]
        if "Color" in meta_data["Pixels"]["Channel"]:
            self.channel_names = [meta_data["Pixels"]["Channel"]["Color"]]

    def read_ome_metadata(self, image_file):
        meta_data = tifffile.xml2dict(image_file.ome_metadata)["OME"]["Image"]
        with suppress(KeyError):
            self.spacing = [
                meta_data["Pixels"][f"PhysicalSize{x}"] * name_to_scalar[meta_data["Pixels"][f"PhysicalSize{x}Unit"]]
                for x in ["Z", "Y", "X"]
            ]
        with suppress(KeyError):
            self.shift = [
                meta_data["Pixels"]["Plane"][0][f"Position{x}"]
                * name_to_scalar[meta_data["Pixels"]["Plane"][0][f"Position{x}Unit"]]
                for x in ["Z", "Y", "X"]
            ]
        self.name = meta_data.get("Name", "")
        self._read_ome_channel_information(meta_data)

    def read_lsm_metadata(self, image_file):
        self.spacing = [image_file.lsm_metadata[f"VoxelSize{x}"] for x in ["Z", "Y", "X"]]
        if "ChannelColors" in image_file.lsm_metadata:
            if "Colors" in image_file.lsm_metadata["ChannelColors"]:
                self.colors = [x[:3] for x in image_file.lsm_metadata["ChannelColors"]["Colors"]]
            if "ColorNames" in image_file.lsm_metadata["ChannelColors"]:
                self.channel_names = image_file.lsm_metadata["ChannelColors"]["ColorNames"]


name_to_scalar = {
    "micron": 10**-6,
    "µm": 10**-6,
    "um": 10**-6,
    "nm": 10**-9,
    "mm": 10**-3,
    "millimeter": 10**-3,
    "pm": 10**-12,
    "picometer": 100**-12,
    "nanometer": 10**-9,
    "\\u00B5m": 10**-6,
    "centimeter": 10**-2,
    "cm": 10**-2,
    "cal": 2.54 * 10**-2,
}  #: dict with known names of scalar to scalar value. Some may be  missed



import os
import sys

from PartSegImage import tifffile_fixes  # noqa: F401
from PartSegImage.channel_class import Channel
from PartSegImage.image import Image
from PartSegImage.image_reader import (
    CziImageReader,
    GenericImageReader,
    ObsepImageReader,
    OifImagReader,
    TiffFileException,
    TiffImageReader,
)
from PartSegImage.image_writer import BaseImageWriter, IMAGEJImageWriter, ImageWriter

__all__ = (
    "BaseImageWriter",
    "Channel",
    "Image",
    "TiffImageReader",
    "IMAGEJImageWriter",
    "ImageWriter",
    "TiffFileException",
    "CziImageReader",
    "OifImagReader",
    "ObsepImageReader",
    "GenericImageReader",
)

if os.path.basename(sys.argv[0]) in ["sphinx-build", "sphinx-build.exe"]:  # pragma: no cover
    for el in __all__:
        globals()[el].__module__ = __name__



from enum import Enum

from local_migrator import register_class


# noinspection NonAsciiCharacters
@register_class(old_paths=["PartSeg.utils.universal_const.Units"])
class Units(Enum):
    mm = 0
    µm = 1  # noqa: PLC2401
    nm = 2
    pm = 3

    def __str__(self):
        return self.name.replace("_", " ")


_UNITS_LIST = ["mm", "µm", "nm", "pm"]
UNIT_SCALE = [10**3, 10**6, 10**9, 10**12]



"""
This module contain utilities to document Register class.
"""

import inspect
from typing import Any, Dict

from sphinx.application import Sphinx
from sphinx.ext.autodoc import ModuleLevelDocumenter

from PartSegCore.algorithm_describe_base import (
    AlgorithmDescribeBase,
    AlgorithmProperty,
    Register,
    get_fields_from_algorithm,
)
from PartSegCore.class_generator import extract_type_info, extract_type_name


# noinspection PyUnusedLocal
def algorithm_parameters_doc(app: Sphinx, what, name: str, obj, options, lines: list):
    if inspect.isclass(obj) and issubclass(obj, AlgorithmDescribeBase) and not inspect.isabstract(obj):
        fields = [x for x in get_fields_from_algorithm(obj) if isinstance(x, AlgorithmProperty)]
        if fields:
            lines.extend(["", "This algorithm has following parameters:", ""])
        for el in fields:
            if el.help_text:
                lines.append(f"- **{el.name}** ({extract_type_name(el.value_type)})- {el.user_name}, {el.help_text}")

            else:
                lines.append(f"- **{el.name}** ({extract_type_name(el.value_type)})- {el.user_name}")


class RegisterDocumenter(ModuleLevelDocumenter):
    objtype = "register_data"
    directivetype = "data"

    @classmethod
    def can_document_member(cls, member: Any, membername: str, isattr: bool, parent: Any) -> bool:
        return isinstance(member, Register)

    def document_members(self, all_members: bool = False) -> None:
        pass

    def add_content(self, more_content: Any, **kwargs) -> None:
        super().add_content(more_content, **kwargs)
        if not isinstance(self.object, Register):
            raise ValueError("Not Register object")
        source = "autogenerated"
        k = 0
        if self.object.methods:
            self.add_line(
                f'Need methods: {", ".join(f"``{x}``" for x in self.object.methods)}',
                source,
                k,
            )

            self.add_line("", source, k + 1)
            k += 2
        if self.object.class_methods:
            self.add_line(
                f'Need class methods: {", ".join(f"``{x}``" for x in self.object.class_methods)}',
                source,
                k,
            )

            self.add_line("", source, k + 1)
            k += 2
        self.add_line("Default content:", source, k)
        self.add_line("", "autogenerated", k + 1)
        k += 2
        for i, (name, val) in enumerate(self.object.items(), k):
            self.add_line(f"- **{name}** - :py:class:`~{extract_type_info(val)[0]}`", "autogenerated", i)
        self.add_line("", "autogenerated", len(self.object) + k)


def setup(app: Sphinx) -> Dict[str, Any]:
    app.connect("autodoc-process-docstring", algorithm_parameters_doc)
    app.add_autodocumenter(RegisterDocumenter)
    return {"version": "0.9", "env_version": 1, "parallel_read_safe": True}



import inspect
from typing import Optional, Union

from qtpy.QtCore import QThread, Signal


class ProgressTread(QThread):
    range_changed = Signal(int, int)
    step_changed = Signal(int)
    error_signal = Signal(Exception)

    def info_function(self, label: str, val: int):
        if label == "max":
            self.range_changed.emit(0, val)
        elif label == "step":
            self.step_changed.emit(val)


class ExecuteFunctionThread(ProgressTread):
    """Generic Thread to execute"""

    def __init__(self, fun, args: Optional[Union[list, tuple]] = None, kwargs: Optional[dict] = None):
        super().__init__()
        self.args = args or []
        self.kwargs = kwargs or {}
        self.function = fun
        self.result = None

    def run(self):
        try:
            if "callback_function" in inspect.signature(self.function).parameters:
                self.result = self.function(*self.args, **self.kwargs, callback_function=self.info_function)
            elif (
                "range_changed" in inspect.signature(self.function).parameters
                and "step_changed" in inspect.signature(self.function).parameters
            ):
                self.result = self.function(
                    *self.args,
                    **self.kwargs,
                    range_changed=self.range_changed.emit,
                    step_changed=self.step_changed.emit,
                )
            else:
                self.result = self.function(*self.args, **self.kwargs)

        except Exception as e:  # pylint: disable=broad-except
            self.error_signal.emit(e)



import dataclasses
import logging

from qtpy.QtCore import QMutex, QThread, Signal

from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm, ROIExtractionResult


class SegmentationThread(QThread):
    """
    Method to run calculation task in separated Thread. This allows to not freeze main window.
    To get info if calculation is done connect to :py:meth:`~.QThread.finished`.
    """

    execution_done = Signal(ROIExtractionResult)
    """
    Signal contains result of segmentation algorithm. Emitted if calculation ends without exception and
    :py:meth:`SegmentationAlgorithm.calculation_run` return not None result.
    """
    progress_signal = Signal(str, int)
    """
    Signal with information about progress. This is proxy for :py:meth:`SegmentationAlgorithm.calculation_run`
    `report_fun` parameter`
    """
    info_signal = Signal(str)
    exception_occurred = Signal(Exception)
    """Signal emitted when some exception occur during calculation. """

    def __init__(self, algorithm: ROIExtractionAlgorithm):
        super().__init__()
        self.finished.connect(self.finished_task)
        self.algorithm = algorithm
        self.clean_later = False
        self.cache = None
        self._image = None
        self._mask = None
        self.mutex = QMutex()
        self.rerun = False, QThread.InheritPriority

    def get_info_text(self):
        """Proxy for :py:meth:`.SegmentationAlgorithm.get_info_text`."""
        return self.algorithm.get_info_text()

    def send_info(self, text, num):
        self.progress_signal.emit(text, num)

    def run(self):
        """the calculation are done here"""
        if self.algorithm.image is None:
            # assertion for running algorithm without image
            logging.error("No image in class {cls_name}", extra={"cls_name": self.algorithm.__class__})
            return
        try:
            segment_data = self.algorithm.calculation_run_wrap(self.send_info)
            if segment_data is not None:
                segment_data = dataclasses.replace(segment_data, file_path=self.algorithm.image.file_path)
        except Exception as e:  # pylint: disable=broad-except
            self.exception_occurred.emit(e)
            return
        if segment_data is None:
            return
        self.execution_done.emit(segment_data)

    def finished_task(self):
        """
        Called on calculation finished. Check if cache is not empty.
        In such case start calculation again with new parameters.
        """
        self.mutex.lock()
        if self.cache is not None:
            args, kwargs = self.cache
            self.algorithm.set_parameters(*args, **kwargs)
            self.cache = None
            self.clean_later = False
        if self._image is not None:
            self.algorithm.set_image(self._image)
            self._image = None
        if self._mask is not None:
            self.algorithm.set_mask(self._mask)
            self._mask = None
        if self.rerun[0]:
            self.rerun = False, QThread.InheritPriority
            super().start(self.rerun[1])
        elif self.clean_later:
            self.algorithm.clean()
            self.clean_later = False
        self.mutex.unlock()

    def clean(self):
        """
        clean cache if thread is running. Call :py:meth:`SegmentationAlgorithm.clean` otherwise. :
        """
        self.mutex.lock()
        if self.isRunning():
            self.clean_later = True
        else:
            self.algorithm.clean()
        self.mutex.unlock()

    def set_parameters(self, *args, **kwargs):
        """
        check if calculation is running.
        If yes then cache parameters until it finish, otherwise call :py:meth:`.SegmentationAlgorithm.set_parameters`
        """
        self.mutex.lock()
        if self.isRunning():
            self.cache = args, kwargs
            self.clean_later = False
        else:
            self.algorithm.set_parameters(*args, **kwargs)
        self.mutex.unlock()

    def set_image(self, image):
        """
        check if calculation is running.
        If yes then cache parameters until it finish, otherwise call :py:meth:`.SegmentationAlgorithm.set_image`
        :param image: image to be set
        """
        self.mutex.lock()
        if self.isRunning():
            self._image = image
        else:
            self.algorithm.set_image(image)
        self.mutex.unlock()

    def set_mask(self, mask):
        """
        check if calculation is running.
        If yes then cache parameters until it finish, otherwise call :py:meth:`.SegmentationAlgorithm.set_mask`
        :param mask: mask to be set
        """
        self.mutex.lock()
        if self.isRunning():
            self._mask = mask
        else:
            self.algorithm.set_mask(mask)
        self.mutex.unlock()

    def start(self, priority: "QThread.Priority" = QThread.InheritPriority):
        """
        If calculation is running remember to restart it with new parameters.

        Otherwise start immediately.
        """
        self.mutex.lock()
        if self.isRunning():
            self.clean_later = False
            self.rerun = True, priority
        else:
            super().start(priority)
        self.mutex.unlock()



import io
import json
import os
import re
import typing
from abc import ABC, abstractmethod
from datetime import datetime
from enum import Enum
from io import BufferedIOBase, BytesIO, IOBase, RawIOBase, StringIO, TextIOBase
from pathlib import Path
from tarfile import TarFile, TarInfo

import imageio
import numpy as np
import pandas as pd
import tifffile
from openpyxl import load_workbook

from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase, AlgorithmProperty
from PartSegCore.json_hooks import partseg_object_hook
from PartSegCore.project_info import ProjectInfoBase
from PartSegCore.utils import EventedDict, ProfileDict, check_loaded_dict, iterate_names
from PartSegImage import ImageWriter
from PartSegImage.image import minimal_dtype


class SegmentationType(Enum):
    analysis = 1
    mask = 2


class WrongFileTypeException(Exception):
    pass


class NotSupportedImage(Exception):
    pass


def check_segmentation_type(tar_file: TarFile) -> SegmentationType:
    names = [x.name for x in tar_file.getmembers()]
    if "algorithm.json" in names:
        return SegmentationType.analysis
    if "metadata.json" in names:
        return SegmentationType.mask
    raise WrongFileTypeException  # pragma: no cover


def get_tarinfo(name, buffer: typing.Union[BytesIO, StringIO]):
    tar_info = TarInfo(name=name)
    buffer.seek(0)
    if isinstance(buffer, BytesIO):
        tar_info.size = len(buffer.getbuffer())
    else:
        tar_info.size = len(buffer.getvalue())
    tar_info.mtime = datetime.now().timestamp()
    return tar_info


class _IOBase(AlgorithmDescribeBase, ABC):
    @classmethod
    def get_name_with_suffix(cls):
        return cls.get_name()

    @classmethod
    def get_extensions(cls) -> typing.List[str]:
        match = re.match(r".*\((.*)\)", cls.get_name())
        if match is None:
            raise ValueError(f"No extensions found in {cls.get_name()}")
        extensions = match[1].split(" ")
        if not all(x.startswith("*.") for x in extensions):
            raise ValueError(f"Error with parsing extensions in {cls.get_name()}")
        return [x[1:] for x in extensions]


class SaveBase(_IOBase, ABC):
    need_functions: typing.ClassVar[typing.List[str]] = [
        "save",
        "get_short_name",
        "get_name_with_suffix",
        "get_default_extension",
        "need_segmentation",
        "need_mask",
    ]

    @classmethod
    @abstractmethod
    def get_short_name(cls):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        """

        :param save_location: location to save, can be buffer
        :param project_info: all project data
        :param parameters: additional parameters for saving method
        :param range_changed: report function for inform about steps num
        :param step_changed: report function for progress
        """
        raise NotImplementedError

    @classmethod
    def need_segmentation(cls):
        return True

    @classmethod
    def need_mask(cls):
        return False

    @classmethod
    def get_default_extension(cls):
        match = re.search(r"\(\*(\.\w+)", cls.get_name_with_suffix())
        return match[1] if match else ""


class LoadBase(_IOBase, ABC):
    need_functions: typing.ClassVar[typing.List[str]] = [
        "load",
        "get_short_name",
        "get_name_with_suffix",
        "number_of_files",
        "get_next_file",
        "partial",
    ]

    @classmethod
    @abstractmethod
    def get_short_name(cls):
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> typing.Union[ProjectInfoBase, typing.List[ProjectInfoBase]]:
        """
        Function for load data

        :param load_locations: list of files to load
        :param range_changed: callback function for inform about number of steps to be done
        :param step_changed:  callback function for report that single step has been done
        :param metadata: additional information needed by function. Like default spacing for load image
        :return: Project info or list of project info
        """
        raise NotImplementedError

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def number_of_files(cls):
        """Number of files required for load method"""
        return 1

    @classmethod
    def get_next_file(cls, file_paths: typing.List[str]):
        return file_paths[0]

    @classmethod
    def partial(cls):
        """Inform that this class load complete data"""
        return False


def load_metadata_base(data: typing.Union[str, Path, typing.TextIO]):
    try:
        if isinstance(data, io.TextIOBase):
            decoded_data = json.load(data, object_hook=partseg_object_hook)
        elif os.path.exists(data):
            with open(data, encoding="utf-8") as ff:
                decoded_data = json.load(ff, object_hook=partseg_object_hook)
        else:
            decoded_data = json.loads(data, object_hook=partseg_object_hook)
    except ValueError as e:  # pragma: no cover
        try:
            decoded_data = json.loads(str(data), object_hook=partseg_object_hook)
        except Exception:
            raise e  # noqa: B904

    return decoded_data


def load_metadata_part(data: typing.Union[str, Path]) -> typing.Tuple[typing.Any, typing.List[typing.Tuple[str, dict]]]:
    """
    Load serialized data. Get valid entries.

    :param data: path to file or string to be decoded.
    :return:
    """
    # TODO extract to function
    data = load_metadata_base(data)
    bad_key = []
    if isinstance(data, typing.MutableMapping) and "__error__" in data:
        bad_key.append(data)
        data = {}
    if isinstance(data, typing.MutableMapping) and not check_loaded_dict(data):
        bad_key.extend((k, data.pop(k)) for k, v in list(data.items()) if not check_loaded_dict(v))
    elif isinstance(data, ProfileDict) and not data.verify_data():
        bad_key = data.pop_errors()
    return data, bad_key


load_matadata_part = load_metadata_part
# backward compatibility


def find_problematic_entries(data: typing.Any) -> typing.List[typing.MutableMapping]:
    """
    Find top nodes with ``"__error__"`` key. If node found
    then its children is not checked.

    :param data: data to be checked
    :return:  top level entries with "__error__" key
    """
    if not isinstance(data, typing.MutableMapping):
        return []
    if "__error__" in data:
        return [data]
    res = []
    for v in data.values():
        res.extend(find_problematic_entries(v))
    return res


def find_problematic_leafs(data: typing.Any) -> typing.List[typing.MutableMapping]:
    """
    Find bottom nodes with ``"__error__"`` key. If any
    children has ``"__error__"`` then such node is not returned.

    :param data: data to be checked.
    :return: bottom level entries with "__error__" key
    """
    if not isinstance(data, typing.MutableMapping):
        return []
    if "__error__" not in data and (not isinstance(data, EventedDict) or len(data) == 0):
        return []
    res = []
    data_to_check = data
    if "__class__" in data and "__values__" in data:
        data_to_check = data["__values__"]
    for data_ in data_to_check.values():
        res.extend(find_problematic_leafs(data_))
    return res or [data]


def proxy_callback(
    range_changed: typing.Callable[[int, int], typing.Any],
    step_changed: typing.Callable[[int], typing.Any],
    text: str,
    val,
):
    if text == "max" and range_changed is not None:
        range_changed(0, val)
    if text == "step" and step_changed is not None:
        step_changed(val)


def open_tar_file(
    file_data: typing.Union[str, Path, TarFile, TextIOBase, BufferedIOBase, RawIOBase, IOBase], mode="r"
) -> typing.Tuple[TarFile, str]:
    """Create tar file from path or buffer. If passed :py:class:`TarFile` then return it."""
    if isinstance(file_data, TarFile):
        tar_file = file_data
        file_path = ""
    elif isinstance(file_data, (str, Path)):
        tar_file = TarFile.open(file_data, mode)
        file_path = str(file_data)
    elif isinstance(file_data, (TextIOBase, BufferedIOBase, RawIOBase, IOBase)):
        tar_file = TarFile.open(fileobj=file_data)
        file_path = ""
    else:
        raise ValueError(f"wrong type of file_data argument: {type(file_data)}")
    return tar_file, file_path


class SaveMaskAsTiff(SaveBase):
    @classmethod
    def get_name(cls):
        return "Mask (*.tiff *.tif)"

    @classmethod
    def get_short_name(cls):
        return "mask_tiff"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def need_mask(cls):
        return True

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: typing.Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        if project_info.image.mask is None and project_info.mask is not None:
            ImageWriter.save_mask(project_info.image.substitute(mask=project_info.mask), save_location)
        else:
            ImageWriter.save_mask(project_info.image, save_location)


def tar_to_buff(tar_file, member_name) -> BytesIO:
    tar_value = tar_file.extractfile(tar_file.getmember(member_name))
    buffer = BytesIO()
    buffer.write(tar_value.read())
    buffer.seek(0)
    return buffer


class SaveScreenshot(SaveBase):
    @classmethod
    def get_short_name(cls):
        return "screenshot"

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: typing.Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        imageio.imsave(save_location, project_info)

    @classmethod
    def get_name(cls) -> str:
        return "Screenshot (*.png *.jpg *.jpeg)"

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return []


class SaveROIAsTIFF(SaveBase):
    @classmethod
    def get_name(cls):
        return "ROI as tiff (*.tiff *.tif)"

    @classmethod
    def get_short_name(cls):
        return "roi_tiff"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: dict,
        range_changed=None,
        step_changed=None,
    ):
        roi = project_info.roi_info.roi
        roi_max = max(project_info.roi_info.bound_info)
        roi = roi.astype(minimal_dtype(roi_max))
        tifffile.imwrite(save_location, roi)


class SaveROIAsNumpy(SaveBase):
    @classmethod
    def get_name(cls):
        return "ROI as numpy (*.npy)"

    @classmethod
    def get_short_name(cls):
        return "ROI_numpy"

    @classmethod
    def get_fields(cls):
        return []

    @classmethod
    def save(
        cls,
        save_location: typing.Union[str, BytesIO, Path],
        project_info,
        parameters: typing.Optional[dict] = None,
        range_changed=None,
        step_changed=None,
    ):
        roi = project_info.roi_info.roi
        roi_max = max(project_info.roi_info.bound_info)
        roi = roi.astype(minimal_dtype(roi_max))
        np.save(save_location, roi)


class PointsInfo(typing.NamedTuple):
    file_path: str
    points: np.ndarray


class LoadPoints(LoadBase):
    @classmethod
    def get_short_name(cls):
        return "point_csv"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> PointsInfo:
        df = pd.read_csv(load_locations[0], delimiter=",", index_col=0)
        return PointsInfo(load_locations[0], df.to_numpy())

    @classmethod
    def get_name(cls) -> str:
        return "Points (*.csv)"

    @classmethod
    def get_fields(cls) -> typing.List[typing.Union[AlgorithmProperty, str]]:
        return ["text"]

    @classmethod
    def partial(cls):
        return True


class LoadPlanJson(LoadBase):
    @classmethod
    def get_short_name(cls):
        return "plan_json"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ):
        from PartSegCore.analysis.calculation_plan import CalculationPlan

        res, err = load_metadata_part(load_locations[0])
        res_dkt = {}
        err_li = []
        for key, value in res.items():
            if isinstance(value, CalculationPlan) and value.is_bad():
                err_li.append(f"Problem with load {value.name} because of {value.get_error_source()}")
            else:
                res_dkt[key] = value
        return res_dkt, err + err_li

    @classmethod
    def get_name(cls) -> str:
        return "Calculation plans (*.json)"


class LoadPlanExcel(LoadBase):
    @classmethod
    def get_short_name(cls):
        return "plan_excel"

    @classmethod
    def load(
        cls,
        load_locations: typing.List[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ):
        data_list, error_list = [], []

        xlsx = load_workbook(filename=load_locations[0], read_only=True)
        try:
            for sheet_name in xlsx.sheetnames:
                if sheet_name.startswith("info"):
                    data = ""
                    index = 2  # skip header
                    while xlsx[sheet_name].cell(row=index, column=2).value:
                        data += xlsx[sheet_name].cell(row=index, column=2).value
                        index += 1

                    try:
                        data, err = load_metadata_part(data)
                        data_list.append(data)
                        error_list.extend(err)
                    except ValueError:  # pragma: no cover
                        error_list.append(f"Cannot load data from: {sheet_name}")
        finally:
            xlsx.close()
        data_dict = {}
        for calc_plan in data_list:
            if calc_plan.is_bad():
                error_list.append(f"Problem with load {calc_plan.name} because of {calc_plan.get_error_source()}")
                continue
            new_name = iterate_names(calc_plan.name, data_dict)
            if new_name is None:  # pragma: no cover
                error_list.append(f"Cannot determine proper name for {calc_plan.name}")
            calc_plan.name = new_name
            data_dict[new_name] = calc_plan
        return data_dict, error_list

    @classmethod
    def get_name(cls) -> str:
        return "Calculation plans from result (*.xlsx)"


IO_LABELS_COLORMAP = "io.labels_colormap_dir"
IO_MASK_METADATA_FILE = "metadata.json"



import os

from napari.layers import Layer

from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_backend import napari_get_settings
from PartSegCore.json_hooks import PartSegEncoder

_SETTINGS = None


class PartSegNapariEncoder(PartSegEncoder):
    def default(self, o):
        if isinstance(o, Layer):
            return o.name
        return super().default(o)


class PartSegNapariSettings(PartSettings):
    json_encoder_class = PartSegNapariEncoder


def get_settings() -> PartSettings:
    global _SETTINGS  # noqa: PLW0603  # pylint: disable=global-statement
    if _SETTINGS is None:
        napari_settings = napari_get_settings()
        if hasattr(napari_settings, "path"):
            save_path = napari_settings.path
        else:
            save_path = os.path.dirname(napari_settings.config_path)
        _SETTINGS = PartSegNapariSettings(os.path.join(save_path, "PartSeg_napari_plugins"))
        _SETTINGS.load()
    return _SETTINGS



import warnings

from PartSegCore.algorithm_describe_base import AlgorithmSelection
from PartSegCore.segmentation.segmentation_algorithm import (
    AutoThresholdAlgorithm,
    CellFromNucleusFlow,
    MorphologicalWatershed,
    SplitImageOnParts,
    ThresholdAlgorithm,
    ThresholdFlowAlgorithm,
    ThresholdPreview,
)


class MaskAlgorithmSelection(
    AlgorithmSelection,
    class_methods=["support_time", "support_z"],
    methods=["set_image", "set_mask", "get_info_text", "calculation_run"],
):
    """Register for segmentation method visible in PartSeg ROI Mask."""


MaskAlgorithmSelection.register(ThresholdAlgorithm)
MaskAlgorithmSelection.register(ThresholdFlowAlgorithm)
MaskAlgorithmSelection.register(ThresholdPreview)
MaskAlgorithmSelection.register(AutoThresholdAlgorithm)
MaskAlgorithmSelection.register(CellFromNucleusFlow)
MaskAlgorithmSelection.register(MorphologicalWatershed)
MaskAlgorithmSelection.register(SplitImageOnParts)


def __getattr__(name):  # pragma: no cover
    if name == "mask_algorithm_dict":
        warnings.warn(
            "mask_algorithm_dict is deprecated. Please use MaskAlgorithmSelection instead",
            category=FutureWarning,
            stacklevel=2,
        )
        return MaskAlgorithmSelection.__register__
    raise AttributeError(f"module {__name__} has no attribute {name}")
